# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
import numpy as np

from karabo.common.states import State
from karabo.native import (
    AccessLevel, AccessMode, Assignment, Bool, ByteArray, Char, Configurable,
    Double, Float, Int8, Int16, Int32, Int64, Node, Slot, String, TypeHash,
    TypeSchema, UInt8, UInt16, UInt32, UInt64, Unit, VectorBool, VectorChar,
    VectorDouble, VectorFloat, VectorHash, VectorInt8, VectorInt16,
    VectorInt32, VectorInt64, VectorString, VectorUInt8, VectorUInt16,
    VectorUInt32, VectorUInt64)
from karabogui.binding.api import (
    BoolBinding, ByteArrayBinding, CharBinding, FloatBinding, HashBinding,
    Int8Binding, Int16Binding, Int32Binding, Int64Binding, NodeBinding,
    SchemaBinding, SlotBinding, StringBinding, Uint8Binding, Uint16Binding,
    Uint32Binding, Uint64Binding, VectorBoolBinding, VectorCharBinding,
    VectorDoubleBinding, VectorFloatBinding, VectorHashBinding,
    VectorInt8Binding, VectorInt16Binding, VectorInt32Binding,
    VectorInt64Binding, VectorStringBinding, VectorUint8Binding,
    VectorUint16Binding, VectorUint32Binding, VectorUint64Binding)


def get_all_props_schema():
    return AllProperties.getClassSchema()


def get_simple_props_schema():
    return SimpleProperties.getClassSchema()


def get_pipeline_schema():
    return PipelineData.getClassSchema()


def get_pipeline_vector_schema():
    return VectorPipelineData.getClassSchema()


def get_simple_schema():
    return Simple.getClassSchema()


def get_slotted_schema():
    return SlottedDevice.getClassSchema()


def get_vectorattr_schema():
    return VectorAttr.getClassSchema()


class Multi(Configurable):
    """`ChoiceOfNodes` and `ListOfNodes` are implemented such that they take a
    single class as their argument, then get their collection of nodes from
    all the classes which inherit from that class. Yeah. I'll take _two_ of
    whatever that guy is having, please.
    """


class _NodeOne(Multi):
    zero = String(defaultValue='First')


class _NodeTwo(Multi):
    one = String(defaultValue='Second')


class OutputNodeInner(Configurable):
    image = Node(Configurable, displayType='Image')


class OutputVectorNodeInner(Configurable):
    vector = VectorDouble()


class OutputVectorNode(Configurable):
    data = Node(OutputVectorNodeInner)


class OutputNode(Configurable):
    data = Node(OutputNodeInner)


class Simple(Configurable):
    """A configurable with a few properties
    """
    foo = Bool(defaultValue=True)
    bar = String(defaultValue="default")
    charlie = Int32(accessMode=AccessMode.READONLY)


class PipelineData(Configurable):
    output = Node(OutputNode, displayType='OutputChannel')


class VectorPipelineData(Configurable):
    output = Node(OutputVectorNode, displayType='OutputChannel')


class SlottedDevice(Configurable):
    state = String(defaultValue=State.ACTIVE)

    @Slot(allowedStates=[State.INIT, State.ACTIVE])
    def callme(self):
        pass


class RowSchema(Configurable):
    start = Float()
    stop = Float()


ALL_PROPERTIES_MAP = {
    'a': BoolBinding, 'b': CharBinding, 'e': FloatBinding, 'f': FloatBinding,
    'g': HashBinding, 'h': Int16Binding, 'i': Int32Binding,
    'j': Int64Binding, 'k': Int8Binding, 'll': SchemaBinding,
    'm': StringBinding, 'n': Uint16Binding, 'o': Uint32Binding,
    'p': Uint64Binding, 'q': Uint8Binding, 'r': VectorBoolBinding,
    's': VectorCharBinding, 'v': VectorDoubleBinding,
    'w': VectorFloatBinding, 'x': VectorHashBinding,
    'y': VectorInt16Binding, 'z': VectorInt32Binding,
    'a1': VectorInt64Binding, 'b1': VectorInt8Binding,
    'c1': VectorStringBinding, 'd1': VectorUint16Binding,
    'e1': VectorUint32Binding, 'f1': VectorUint64Binding,
    'g1': VectorUint8Binding, 'h1': NodeBinding,
    'k1': SlotBinding, 'mm': ByteArrayBinding,
}


class AllProperties(Configurable):
    """A `Configurable` with every type of property
    """
    a = Bool(defaultValue=True, requiredAccessLevel=AccessLevel.EXPERT)
    b = Char(defaultValue='c')
    e = Double(unitSymbol=Unit.METER)
    f = Float()
    g = TypeHash()
    h = Int16()
    i = Int32(accessMode=AccessMode.READONLY, assignment=Assignment.INTERNAL)
    j = Int64(unitSymbol=Unit.METER)
    k = Int8()
    ll = TypeSchema()
    m = String(options=['foo', 'bar', 'baz', 'qux'])
    n = UInt16()
    o = UInt32()
    p = UInt64()
    q = UInt8()
    r = VectorBool()
    s = VectorChar()
    v = VectorDouble()
    w = VectorFloat()
    x = VectorHash(rows=RowSchema)
    y = VectorInt16()
    z = VectorInt32()
    a1 = VectorInt64()
    b1 = VectorInt8()
    c1 = VectorString()
    d1 = VectorUInt16()
    e1 = VectorUInt32()
    f1 = VectorUInt64()
    g1 = VectorUInt8()
    h1 = Node(_NodeOne)
    mm = ByteArray()

    @Slot(allowedStates=[State.INTERLOCKED, State.ACTIVE])
    def k1(self):
        pass


class VectorAttr(Configurable):
    vec = VectorBool(defaultValue=np.array([True, True]))


class SimpleProperties(Configurable):
    """A `Configurable` with every type of property
    """
    boolProperty = Bool(defaultValue=True)
    falseProperty = Bool(defaultValue=False)
    doubleProperty = Double(unitSymbol=Unit.METER)
    floatProperty = Float(accessMode=AccessMode.READONLY)
    intProperty = Int16(accessMode=AccessMode.READONLY)
    stringProperty = String(options=['foo', 'bar', 'baz', 'qux'])
    vectorProperty = VectorDouble()
    vectorStringProperty = VectorString()
    table = VectorHash(Simple)
    node = Node(Simple)
    internal = Bool(defaultValue=True,
                    assignment=Assignment.INTERNAL)

    @Slot(allowedStates=[State.INTERLOCKED, State.ACTIVE])
    def anySlot(self):
        pass



# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
from qtpy import uic
from qtpy.QtCore import Slot
from qtpy.QtWidgets import QDialog
from traits.api import TraitError, Undefined

from karabogui.controllers.validators import SimpleValidator

from .utils import get_dialog_ui

_ALARM_KEYS = ["alarmLow", "warnLow", "warnHigh", "alarmHigh"]

LABEL_NOTE = ("Note: In order to deactivate an alarm setting, please leave "
              "the edit field empty.")


class AlarmDialog(QDialog):

    def __init__(self, binding, config, parent=None):
        super().__init__(parent=parent)
        uic.loadUi(get_dialog_ui("alarm_dialog.ui"), self)
        for key, value in config.items():
            if key not in _ALARM_KEYS:
                continue
            value = "" if value is Undefined else str(value)
            widget = getattr(self, f"edit_{key}")
            widget.setText(value)
            widget.textChanged.connect(self.validate_the_order)

        self.binding = binding
        self.validator = SimpleValidator(binding)
        self.edit_warnLow.setValidator(self.validator)
        self.edit_warnHigh.setValidator(self.validator)
        self.edit_alarmLow.setValidator(self.validator)
        self.edit_alarmHigh.setValidator(self.validator)

    @property
    def values(self):
        """Get the alarm configuration in order"""
        traits = self._get_traits()

        active = {}
        for key, value in traits.items():
            if value is not Undefined:
                active[key] = value

        # Reorder in case of wrong input, but only the active ones
        active = {k: v for k, v in zip(active.keys(), sorted(active.values()))}
        traits.update(active)

        return traits

    def _get_traits(self):
        traits = {}
        for key in _ALARM_KEYS:
            widget = getattr(self, f"edit_{key}")
            value = widget.text()
            value = (Undefined if not value or not widget.hasAcceptableInput()
                     else self.trait_value(value))
            traits[key] = value
        return traits

    def trait_value(self, value):
        try:
            return self.binding.validate_trait("value", value)
        except TraitError:
            return Undefined

    @Slot()
    def validate_the_order(self):
        """
        Validate the order of all the limit values. It should always be
           alarmLow < warnLow < warnHigh <  alarmHigh.

        Empty values are ignored. For example if "warnLow" is not defined
        then, the condition is
           alarmLow < warnHigh <  alarmHigh.

        A message is displayed in red color when the values do not follow
        the order.
        """
        traits = self._get_traits()
        actives = {k: v for k, v in traits.items() if v is not Undefined}
        values = list(actives.values())
        valid_order = values == sorted(values, key=lambda x: float(x))
        if not valid_order:
            order = " < ".join(actives.keys())
            message = f"Please make sure the values are in the order: {order}"
            style = "QLabel#label_note {color:red;}"
        else:
            message = LABEL_NOTE
            style = "QLabel#label_note {}"
        self.label_note.setText(message)
        self.label_note.setStyleSheet(style)



# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
from qtpy.QtWidgets import QFrame, QHBoxLayout, QLabel, QSizePolicy


class UnitLabelWrapper(QFrame):
    """NOTE: We are not deriving a class from QWidget here, because doing so
    breaks the usage of stylesheets elsewhere in the GUI (configurator).
    The `update_label` function is attached to the returned widget as a
    workaround...
    """

    def __init__(self, widget, parent=None):
        super().__init__(parent)

        # Store internal widget reference
        self._internal_widget = widget

        # Add label
        self.label = label = QLabel(self)
        label.setSizePolicy(QSizePolicy.Fixed, QSizePolicy.Preferred)

        # Add and populate layout
        layout = QHBoxLayout(self)
        layout.setSizeConstraint(QHBoxLayout.SetNoConstraint)
        layout.setSpacing(0)
        layout.setContentsMargins(0, 0, 0, 0)
        layout.addWidget(widget)
        layout.addWidget(label)

    def update_unit_label(self, proxy):
        """Update the label according to the unit_label of the `proxy`"""
        if proxy.binding is None:
            return

        unit_label = proxy.binding.unit_label
        self.label.setVisible(unit_label != "")
        self.label.setText(unit_label)

    def update_label(self, proxy):
        # Add an `update_label` "method" for keeping things synced
        if proxy.binding is None:
            return

        unit_label = proxy.binding.unit_label
        if unit_label != self.label.text():
            self.label.setText(unit_label)

    def setFont(self, font):
        super().setFont(font)
        self._internal_widget.setFont(font)
        self.label.setFont(font)


def add_unit_label(proxy, widget, parent=None):
    """Add a unit label to a widget, if available."""

    wrapper_widget = UnitLabelWrapper(widget, parent=parent)
    wrapper_widget.update_label(proxy)
    return wrapper_widget



#############################################################################
# Author: <dennis.goeries@xfel.eu>
# Created on September 16, 2018
# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
#############################################################################
from collections import defaultdict
from typing import Any

from qtpy.QtCore import QObject, QSettings

from .util import decrypt, encrypt


def convert_value(value, dtype):
    """Convert a value to a QSettings value. This is mainly done for booleans
    """
    if dtype is bool:
        return str(value).lower()
    return value


class Item:
    __slots__ = (
        "name", "default", "q_set", "group", "path", "editable", "dtype")

    def __init__(self, default: Any = None,
                 q_set: bool = False,
                 group: str | None = None,
                 editable: bool = False,
                 dtype: int | float | bool | None = None) -> None:
        self.name: str = ""
        self.default = default
        self.q_set = q_set
        self.group = group
        self.path: str = ""
        self.editable = editable
        self.dtype = dtype

    def __get__(self, instance: "Item", owner: type["Item"]) -> Any:
        if instance is None:
            return self
        else:
            return instance.__dict__.get(self.name, self.default)

    def __set__(self, instance: "Item", value: Any) -> None:
        instance.__dict__[self.name] = value
        if self.q_set:
            QSettings().setValue(self.path, convert_value(value, self.dtype))

    def __set_name__(self, owner: type["Item"], name: str) -> None:
        self.name = name
        self.path = f"{self.group}/{self.name}"
        if self.q_set:
            self.default = self.get_shared_value()

    def get_shared_value(self) -> Any:
        """Initialize a value from `QSettings` file"""
        value = QSettings().value(self.path)
        if not value or self.dtype is None:
            return value or self.default
        # Actively stored values
        if self.dtype is bool:
            return str(value).lower() == "true"
        elif self.dtype is int:
            return int(value)
        elif self.dtype is float:
            return float(value)
        return value

    def toDict(self) -> dict:
        return {name: getattr(self, name) for name in
                ["editable", "name", "dtype"]}

    def erase(self) -> None:
        QSettings().remove(self.path)

    def __str__(self) -> str:
        return self.name


class SharedItem(Item):
    """A SharedItem in the Configuration is by default storing into QSettings
    """

    def __init__(self, group: str | None = None,
                 editable: bool = False,
                 dtype: int | float | bool | None = None) -> None:
        super().__init__(default=None, q_set=True, group=group,
                         editable=editable, dtype=dtype)

    def __get__(self, instance: "SharedItem",
                owner: type["SharedItem"]) -> Any:
        if instance is None:
            return self
        value = self.get_shared_value()
        return decrypt(value, self.group) if value is not None else value

    def __set__(self, instance: "SharedItem", value: Any) -> None:
        value = convert_value(value, self.dtype)
        if value is not None:
            value = encrypt(value, self.group)
        QSettings().setValue(self.path, value)

    def __set_name__(self, owner: type["SharedItem"], name: str) -> None:
        self.name = name
        self.path = f"{self.group}/{self.name}"
        self.default = None


AUTHENTICATION = "authentication"
NETWORK = "network"
PROJECT = "project"
BACKBONE = "backbone"
DIRECTORIES = "dir"
USER = "user"


class Configuration(QObject):
    """The main dispatch point for Karabo Configurations and QSettings.

    An Item is constructed via the descriptor protocol.

    pool = Item(default="fish", q_set=False, group="POOL")

    - The first parameter ``default`` provides - as the name says - the
      default value. This value might be overwritten on initialization if
      there is a presetting in QSettings.

    - The ``qset`` parameter activates the QSetting functionality. Items with
      this parameter set to True, store their value in a QSetting file and load
      this on the next GUI startup and overwrite an eventual default value.

    - The ``editable`` parameter defines if this item is user editable

    - The ``group`` parameter defines an additional string for sorting and
      grabbing the Items for a model view.

    - The ``dtype`` parameter defines the data dtype. If a dtype is specified,
      a default value must be provided.

    NOTE: Every parameter has a cache principle, meaning that if an additional
          client is opened on the same machine, both clients do not interact
          via QSettings on runtime.

    NOTE: Not every parameter is supposed to have a QSetting.

          - db tokens or other sensible data which should not leave the GUI.
          - parameters, such as broker topic, which might not be send always
            as they will take a default value from QSettings next time, which
            might be wrong.
    """
    broker_topic = Item(q_set=False, group=BACKBONE)

    # ----------------------------------------------
    # Last directories stored and used

    data_dir = Item(q_set=True, group=DIRECTORIES)

    # ----------------------------------------------
    # MainWindow, Wizard and DPI

    wizard = Item(default=True, q_set=True, group=USER, editable=True,
                  dtype=bool)
    check_updates = Item(default=True, q_set=True, group=USER, editable=True,
                         dtype=bool)
    highDPI = Item(default=True, q_set=True, group=USER, editable=False,
                   dtype=bool)
    main_geometry = Item(q_set=True, group=USER)
    development = Item(default=False, q_set=False, group=USER, dtype=bool)
    project_sort_column = Item(default=0, q_set=True, group=USER, dtype=int)
    project_sort_order = Item(default=0, q_set=True, group=USER, dtype=int)
    logbook_stream = Item(q_set=True, group=USER)
    logbook_topic = Item(q_set=True, group=USER)
    logbook_header_style = Item(q_set=True, group=USER)

    # ----------------------------------------------
    # Project db interface

    domain = Item(q_set=True, group=PROJECT)

    # ----------------------------------------------
    # GUI Server network connection

    username = Item(default="operator", q_set=True, group=NETWORK)
    access_level = Item(default="operator", q_set=True, group=NETWORK)
    gui_servers = Item(default=[], q_set=True, group=NETWORK)
    macro_development = Item(q_set=True, group=NETWORK)

    # ----------------------------------------------
    # Shared authentication interface

    refresh_token_user = SharedItem(group=AUTHENTICATION)
    refresh_token = SharedItem(group=AUTHENTICATION)

    def __new__(cls, *args, **kwargs):
        instance = super().__new__(cls, *args, **kwargs)
        # Memory filling avoiding __setattr__!
        memory = [getattr(cls, attr).name for attr in dir(cls)
                  if isinstance(getattr(cls, attr), Item)]
        instance.__dict__["_memory"] = memory
        return instance

    def __init__(self, parent=None):
        super().__init__(parent)

    def __setitem__(self, key, value):
        if key not in self._memory:
            raise KeyError(f"Configuration key is not valid: {key}")
        super().__setattr__(key, value)

    def __getitem__(self, item):
        if item not in self._memory:
            raise KeyError(f"Configuration key is not valid: {item}")
        return super().__getattribute__(item)

    def __contains__(self, key):
        return key in self._memory

    def __iter__(self):
        for key in self._memory:
            yield key, super().__getattribute__(key)

    def items(self):
        for key in self._memory:
            yield key, super().__getattribute__(key)

    def values(self):
        for key in self._memory:
            yield super().__getattribute__(key)

    def keys(self):
        yield from self._memory

    def groups(self):
        """Return a dictionary with groups as keys and config keys as values
        """
        ret = defaultdict(list)
        for key in self._memory:
            group = getattr(self.__class__, key).group
            if group == AUTHENTICATION:
                continue
            ret[group].append(getattr(self.__class__, key))
            ret[group] = sorted(ret[group], key=lambda i: i.name)
        return dict(sorted(ret.items()))

    def __delitem__(self, key):
        """Convenient shortcut to erase configuration information"""
        assert key in self._memory
        item = getattr(self.__class__, key)
        item.erase()

    def info(self):
        """Return the relevant user meta data information from the config

        This method only accounts `USER` group information.
        """
        ret = {USER: {}}
        for key in self._memory:
            group = getattr(self.__class__, key).group
            if group != USER:
                continue
            ret[group][key] = getattr(self, key)
        return ret

    def __len__(self):
        return len(self._memory)

    def __setattr__(self, key, value):
        if key not in self._memory:
            raise KeyError(f"Configuration key is not valid: {key}")
        super().__setattr__(key, value)

    def __repr__(self):
        names = "\n".join("\t" + attr + ": " + repr(getattr(self, attr))
                          for attr in self._memory)
        return f"{{\n{names}\n}}"



# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
import numpy as np
from scipy.stats import norm


def gaussian_fit(x, height, x0, sigma, offset):
    return height * np.exp(-0.5 * ((x - x0) / sigma) ** 2) + offset


def normal_cdf(x, mu, sigma):
    """
    :param mu: Mean of the normal distribution.
    :type mu: float
    :param sigma: Standard deviation of the normal distribution.
    :type sigma: float
    """
    return norm.cdf(x, loc=mu, scale=sigma)


def linear_function_fit(x, m, b):
    """
    :param m: Slope of the linear function
    :type m: Float
    :param b: Intercept of the linear function
    :type b: Float
    """
    return m * x + b


def sqsech(x, height, x0, sx, offset):
    """Returns a squared hyperbolic secant  with the given parameters.
    The hyperbolic secant curve is in the form
    height / ((np.cosh((x0 - x) / sx)) ** 2)
    """

    return height / ((np.cosh((x0 - x) / sx)) ** 2) + offset


def guess_initial_parameters(x, y):
    """
    Generate initial parameter for 'curve_fit' function for gaussian and
    sech squared fitting
    """
    height_guess = max(y)
    x0_guess = x[np.argmax(y)]
    sigma_guess = np.std(x)
    offset_guess = np.min(y)
    return [height_guess, x0_guess, sigma_guess, offset_guess]



# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
import numpy as np
import pytest

from karabo.common.scenemodel.api import VectorScatterGraphModel
from karabo.native import Configurable, VectorFloat, VectorInt32
from karabogui.binding.proxy import PropertyProxy
from karabogui.testing import get_class_property_proxy, set_proxy_value

from ..vector_scatter_graph import DisplayVectorScatterGraph


class Object(Configurable):
    x = VectorFloat(defaultValue=[2.3, 4.5, 7.9])
    y = VectorInt32(defaultValue=[1, 2, 3])


@pytest.fixture
def vector_scatter_setup(gui_app):
    schema = Object.getClassSchema()
    proxy = get_class_property_proxy(schema, "x")
    device = proxy.root_proxy
    prop_proxy = PropertyProxy(root_proxy=device, path="y")
    controller = DisplayVectorScatterGraph(proxy=proxy)
    controller.create(None)
    controller.visualize_additional_property(prop_proxy)
    assert controller.widget is not None

    yield controller, proxy, prop_proxy

    controller.destroy()
    assert controller.widget is None


def test_vector_scatter_graph_basics(vector_scatter_setup):
    controller, proxy, prop_proxy = vector_scatter_setup
    set_proxy_value(proxy, "x", [2.3, 4.5, 7.9])
    set_proxy_value(prop_proxy, "y", [1, 2, 3])
    curve = controller._plot
    x, y = curve.getData()
    np.testing.assert_array_almost_equal(x, [2.3, 4.5, 7.9])
    np.testing.assert_array_almost_equal(y, [1, 2, 3])

    # y is different
    set_proxy_value(proxy, "x", [2.3, 4.5, 7.9])
    set_proxy_value(prop_proxy, "y", [1, 2, 3, 4])
    x, y = curve.getData()
    np.testing.assert_array_almost_equal(x, [2.3, 4.5, 7.9])
    np.testing.assert_array_almost_equal(y, [1, 2, 3])

    # x is different
    set_proxy_value(proxy, "x", [2.3, 4.5])
    x, y = curve.getData()
    np.testing.assert_array_almost_equal(x, [2.3, 4.5, 7.9])
    np.testing.assert_array_almost_equal(y, [1, 2, 3])
    # No replot happened, y provides the update, we have size 2 on x
    set_proxy_value(prop_proxy, "y", [1, 2, 3, 4])
    x, y = curve.getData()
    np.testing.assert_array_almost_equal(x, [2.3, 4.5])
    np.testing.assert_array_almost_equal(y, [1, 2])


def test_vector_scatter_graph_pointsize(vector_scatter_setup, mocker):
    _, proxy, prop_proxy = vector_scatter_setup
    controller = DisplayVectorScatterGraph(proxy=proxy,
                                           model=VectorScatterGraphModel())
    controller.create(None)
    action = controller.widget.actions()[10]
    assert action.text() == "Point Size"

    dsym = ("karabogui.controllers.display."
            "vector_scatter_graph.QInputDialog")
    QInputDialog = mocker.patch(dsym)
    QInputDialog.getDouble.return_value = 2.7, True
    action.trigger()
    assert controller.model.psize == 2.7
    assert controller._plot.opts["size"] == 2.7

    controller.destroy()


def test_removal_property(vector_scatter_setup):
    _, proxy, prop_proxy = vector_scatter_setup
    controller = DisplayVectorScatterGraph(proxy=proxy,
                                           model=VectorScatterGraphModel())
    controller.create(None)
    controller.visualize_additional_property(prop_proxy)
    set_proxy_value(proxy, "x", [2.3, 4.5, 7.9])
    set_proxy_value(prop_proxy, "y", [1, 2, 3])

    curve = controller._plot
    x, y = curve.getData()
    np.testing.assert_array_almost_equal(x, [2.3, 4.5, 7.9])
    np.testing.assert_array_almost_equal(y, [1, 2, 3])

    assert controller.remove_additional_property(prop_proxy)
    assert controller._y_proxy is None
    # Removing more does not hurt
    assert not controller.remove_additional_property(prop_proxy)



# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
from karabo.common.api import KARABO_SCHEMA_DISPLAY_TYPE
from karabo.common.display_types import (
    KARABO_SCHEMA_DISPLAY_TYPE_ALARM, KARABO_SCHEMA_DISPLAY_TYPE_BIN,
    KARABO_SCHEMA_DISPLAY_TYPE_HEX, KARABO_SCHEMA_DISPLAY_TYPE_STATE)
from karabo.common.scenemodel.api import (
    AlarmGraphModel, FilterTableElementModel, HistoricTextModel,
    StateGraphModel, TrendGraphModel, VectorGraphModel, WebCamGraphModel,
    get_alarm_graph_scene, get_image_scene, get_state_graph_scene,
    get_text_history_scene, get_trendline_scene, get_vector_scene)
from karabogui.binding.api import (
    BoolBinding, FloatBinding, ImageBinding, IntBinding, NDArrayBinding,
    NodeBinding, PipelineOutputBinding, StringBinding, VectorHashBinding,
    VectorNumberBinding, VectorStringBinding)


def _iter_binding(node, base=""):
    namespace = node.value
    base = base + "." if base else ""
    for name in namespace:
        subname = base + name
        subnode = getattr(namespace, name)
        if isinstance(subnode, NodeBinding):
            yield from _iter_binding(subnode, base=subname)
        yield subname, subnode


def get_generic_scene(proxy, include_images=True):
    """Return a generic scene for a proxy binding

    :param proxy: property proxy
    :param include_images: If image widget are included in the retrieval.
                           The default is `True`
    """
    binding = getattr(proxy, 'binding', None)
    if binding is None:
        return

    instance_id = proxy.root_proxy.device_id
    displayType = binding.attributes.get(KARABO_SCHEMA_DISPLAY_TYPE, '')
    if displayType == KARABO_SCHEMA_DISPLAY_TYPE_STATE:
        path = proxy.path
        return get_state_graph_scene(instance_id, path)
    elif displayType == KARABO_SCHEMA_DISPLAY_TYPE_ALARM:
        path = proxy.path
        return get_alarm_graph_scene(instance_id, path)

    if isinstance(binding, StringBinding):
        path = proxy.path
        return get_text_history_scene(instance_id, path)

    elif isinstance(binding, (BoolBinding, FloatBinding, IntBinding)):
        path = proxy.path
        return get_trendline_scene(instance_id, path)

    elif isinstance(binding, (NDArrayBinding, VectorNumberBinding)):
        path = proxy.path
        return get_vector_scene(instance_id, path)

    elif include_images and isinstance(binding, PipelineOutputBinding):
        for path, node in _iter_binding(binding, base=proxy.path):
            if isinstance(node, ImageBinding):
                return get_image_scene(instance_id, path)

    elif include_images and isinstance(binding, ImageBinding):
        return get_image_scene(instance_id, proxy.path)

    return None


def _get_plot_attributes(proxy):
    """Retrieve the basic plot model properties according to `proxy`"""
    binding = proxy.binding
    y_units = binding.unit_label if binding is not None else ""
    return {
        "y_label": proxy.path,
        "y_units": y_units,
        "width": 600,
        "height": 350,
        "x_grid": True,
        "y_grid": True,
        "keys": [proxy.key]
    }


def _get_widget_attributes(key):
    """Retrieve the basic image model properties according to `proxy`"""
    return {
        "width": 800,
        "height": 600,
        "keys": [key]
    }


def _get_table_attributes(key):
    """Retrieve the basic image model properties according to `proxy`"""
    return {
        "width": 800,
        "height": 600,
        "showFilterKeyColumn": True,
        "keys": [key]
    }


def get_property_proxy_model(proxy, include_images=True):
    """Return a generic model instance for a proxy binding

    :param proxy: property proxy
    :param include_images: If image widget are included in the retrieval.
                           The default is `True`
    """
    binding = getattr(proxy, "binding", None)
    if binding is None:
        return

    displayType = binding.attributes.get(
        KARABO_SCHEMA_DISPLAY_TYPE, "").split("|")[0]
    if displayType == KARABO_SCHEMA_DISPLAY_TYPE_STATE:
        return StateGraphModel(**_get_plot_attributes(proxy))
    elif displayType == KARABO_SCHEMA_DISPLAY_TYPE_ALARM:
        return AlarmGraphModel(**_get_plot_attributes(proxy))
    elif displayType in (
            KARABO_SCHEMA_DISPLAY_TYPE_BIN, KARABO_SCHEMA_DISPLAY_TYPE_HEX):
        return HistoricTextModel(**_get_widget_attributes(proxy.key))

    if isinstance(binding, (StringBinding, VectorStringBinding)):
        return HistoricTextModel(**_get_widget_attributes(proxy.key))

    elif isinstance(binding, VectorHashBinding):
        return FilterTableElementModel(**_get_table_attributes(proxy.key))

    elif isinstance(binding, (BoolBinding, FloatBinding, IntBinding)):
        return TrendGraphModel(**_get_plot_attributes(proxy))

    elif isinstance(binding, (NDArrayBinding, VectorNumberBinding)):
        return VectorGraphModel(**_get_plot_attributes(proxy))

    elif include_images and isinstance(binding, PipelineOutputBinding):
        for path, node in _iter_binding(binding, base=proxy.path):
            if isinstance(node, ImageBinding):
                path = f"{proxy.root_proxy.device_id}.{path}"
                return WebCamGraphModel(**_get_widget_attributes(path))

    elif include_images and isinstance(binding, ImageBinding):
        return WebCamGraphModel(**_get_widget_attributes(proxy.key))

    return None



# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
from qtpy.QtGui import QValidator

from karabo.common.const import KARABO_SCHEMA_MAX_EXC, KARABO_SCHEMA_MIN_INC
from karabogui.binding.api import (
    FloatBinding, Uint8Binding, VectorFloatBinding, VectorUint8Binding)

from ..validators import BindingValidator, ListValidator, SimpleValidator


def test_generic_binding_validator():
    binding = FloatBinding()
    validator = BindingValidator(binding=binding)
    result, _, _ = validator.validate("", None)
    assert result == QValidator.Intermediate
    result, _, _ = validator.validate("2.12", None)
    assert result == QValidator.Acceptable
    result, _, _ = validator.validate("2....", None)
    assert result == QValidator.Intermediate
    result, _, _ = validator.validate("2.", None)
    assert result == QValidator.Acceptable
    # A very generic validator case. Can still cast with traits ...
    result, _, _ = validator.validate(" 2.", None)
    assert result == QValidator.Acceptable
    result, _, _ = validator.validate(" ", None)
    assert result == QValidator.Intermediate
    result, _, _ = validator.validate("2 ", None)
    assert result == QValidator.Acceptable

    binding = VectorFloatBinding()
    validator = BindingValidator(binding=binding)
    result, _, _ = validator.validate([], None)
    assert result == QValidator.Acceptable
    # float and integer
    result, _, _ = validator.validate([2.2, 1], None)
    assert result == QValidator.Acceptable
    # float and string that can be casted.
    result, _, _ = validator.validate([2, "2"], None)
    assert result == QValidator.Acceptable
    # Even bools are casted by traits ...
    result, _, _ = validator.validate([True, False], None)
    assert result == QValidator.Acceptable


def test_list_binding_unsigned_validator():
    binding = VectorUint8Binding()
    validator = ListValidator(binding=binding)
    result, _, _ = validator.validate("", None)
    assert result == QValidator.Acceptable

    result, _, _ = validator.validate("1", None)
    assert result == QValidator.Acceptable

    result, _, _ = validator.validate("0", None)
    assert result == QValidator.Acceptable

    result, _, _ = validator.validate("0,2,1,12,45,12", None)
    assert result == QValidator.Acceptable

    result, _, _ = validator.validate("+", None)
    assert result == QValidator.Intermediate

    result, _, _ = validator.validate("-", None)
    assert result == QValidator.Invalid

    # But not boolean
    result, _, _ = validator.validate("Fals", None)
    assert result == QValidator.Invalid
    # Floating input
    result, _, _ = validator.validate("0.", None)
    assert result == QValidator.Invalid

    result, _, _ = validator.validate("0.0", None)
    assert result == QValidator.Invalid

    # Sign at the end
    result, _, _ = validator.validate("0-", None)
    assert result == QValidator.Invalid

    result, _, _ = validator.validate("0.0-", None)
    assert result == QValidator.Invalid

    result, _, _ = validator.validate("0.0e", None)
    assert result == QValidator.Invalid


def test_list_binding_float_validator():
    binding = VectorFloatBinding()
    validator = ListValidator(binding=binding)
    result, _, _ = validator.validate("", None)
    assert result == QValidator.Acceptable

    # A double regex can have integer input
    result, _, _ = validator.validate("1", None)
    assert result == QValidator.Acceptable
    result, _, _ = validator.validate("0", None)
    assert result == QValidator.Acceptable

    # But not boolean
    result, _, _ = validator.validate("Fals", None)
    assert result == QValidator.Invalid

    # Floating input
    result, _, _ = validator.validate("0.", None)
    assert result == QValidator.Intermediate

    result, _, _ = validator.validate("0.0", None)
    assert result == QValidator.Acceptable

    # Sign at the end
    result, _, _ = validator.validate("0-", None)
    assert result == QValidator.Intermediate

    result, _, _ = validator.validate("0.0-", None)
    assert result == QValidator.Intermediate

    result, _, _ = validator.validate("0.0+", None)
    assert result == QValidator.Intermediate

    result, _, _ = validator.validate("0.0e", None)
    assert result == QValidator.Intermediate

    result, _, _ = validator.validate("0.0e-7", None)
    assert result == QValidator.Acceptable

    result, _, _ = validator.validate("+0000.1223", None)
    assert result == QValidator.Acceptable

    result, _, _ = validator.validate("+m", None)
    assert result == QValidator.Invalid

    result, _, _ = validator.validate("+e", None)
    assert result == QValidator.Intermediate

    result, _, _ = validator.validate("xfel", None)
    assert result == QValidator.Invalid


def test_list_validator_size():
    binding = VectorFloatBinding(attributes={"minSize": 1, "maxSize": 3})
    validator = ListValidator(binding=binding)

    # We are not allowed to be empty!
    result, _, _ = validator.validate("", None)
    assert result == QValidator.Intermediate

    # Accepted in limits
    result, _, _ = validator.validate("1", None)
    assert result == QValidator.Acceptable
    result, _, _ = validator.validate("1,", None)
    assert result == QValidator.Intermediate

    result, _, _ = validator.validate("1,2,3", None)
    assert result == QValidator.Acceptable

    # Out of limits
    result, _, _ = validator.validate("1,2,3,4", None)
    assert result == QValidator.Intermediate


def test_list_integer_limits():
    binding = VectorUint8Binding()
    validator = ListValidator(binding=binding)

    result, _, _ = validator.validate("1", None)
    assert result == QValidator.Acceptable

    result, _, _ = validator.validate("-1", None)
    assert result == QValidator.Invalid

    # Overflowing are intermediate for lists
    result, _, _ = validator.validate("2550", None)
    assert result == QValidator.Intermediate

    result, _, _ = validator.validate("255", None)
    assert result == QValidator.Acceptable


def test_simple_validator():
    binding = Uint8Binding()
    validator = SimpleValidator(binding=binding)
    result, _, _ = validator.validate("", None)
    assert result == QValidator.Intermediate
    result, _, _ = validator.validate("1", None)
    assert result == QValidator.Acceptable
    result, _, _ = validator.validate("-1", None)
    assert result == QValidator.Invalid
    result, _, _ = validator.validate("-", None)
    assert result == QValidator.Intermediate
    result, _, _ = validator.validate("0", None)
    assert result == QValidator.Acceptable
    result, _, _ = validator.validate("255", None)
    assert result == QValidator.Acceptable
    result, _, _ = validator.validate("256", None)
    assert result == QValidator.Invalid
    result, _, _ = validator.validate("0001", None)
    assert result == QValidator.Intermediate

    # Test with binding attributes
    attributes = {KARABO_SCHEMA_MIN_INC: 2, KARABO_SCHEMA_MAX_EXC: 128}
    binding = Uint8Binding(attributes=attributes)
    validator = SimpleValidator(binding=binding)
    result, _, _ = validator.validate("", None)
    assert result == QValidator.Intermediate
    result, _, _ = validator.validate("1", None)
    assert result == QValidator.Intermediate
    result, _, _ = validator.validate("2", None)
    assert result == QValidator.Acceptable
    result, _, _ = validator.validate("-1", None)
    assert result == QValidator.Invalid
    result, _, _ = validator.validate("-", None)
    assert result == QValidator.Intermediate
    result, _, _ = validator.validate("0", None)
    assert result == QValidator.Intermediate
    result, _, _ = validator.validate("255", None)
    assert result == QValidator.Intermediate
    result, _, _ = validator.validate("128", None)
    assert result == QValidator.Intermediate
    result, _, _ = validator.validate("127", None)
    assert result == QValidator.Acceptable



# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
# flake8: noqa
from karabo.native import Hash

from ..util import decrypt, encrypt, get_error_message, realign_topo_hash

CPP_SLOT_ERROR_MESSAGE = """
Request to execute 'execute2' on device 'Macro-test1-ec6d56fe-a5ca-445f-ad45-a295c7e16450-Test' failed, details:
1. Exception =====>  {
    Exception Type....:  Remote Exception from Macro-test1-ec6d56fe-a5ca-445f-ad45-a295c7e16450-Test
    Message...........:  Traceback (most recent call last):
  File "/scratch/xctrl/karabo/extern/lib/python3.6/site-packages/karabo/native/data/hash.py", line 591, in wrapper
    ret = yield from coro
  File "/scratch/xctrl/karabo/extern/lib/python3.6/site-packages/karabo/middlelayer_api/eventloop.py", line 644, in run_coroutine_or_thread
    return (yield from future)
  File "/scratch/xctrl/karabo/extern/lib/python3.6/site-packages/karabo/middlelayer_api/eventloop.py", line 630, in thread
    ret = f(*args, **kwargs)
  File "/scratch/xctrl/karabo/extern/lib/python3.6/site-packages/karabo/native/data/hash.py", line 559, in wrapper
    return self.method(device)
  File "/scratch/xctrl/karabo/extern/lib/python3.6/site-packages/karabo/middlelayer_api/macro.py", line 82, in wrapper
    return themethod(device)
  File "test1", line 20, in execute2
  File "test1", line 15, in do2
  File "/scratch/xctrl/karabo/extern/lib/python3.6/site-packages/karabo/middlelayer_api/proxy.py", line 164, in method
    return myself._parent._callSlot(self, *args, **kwargs)
  File "/scratch/xctrl/karabo/extern/lib/python3.6/site-packages/karabo/middlelayer_api/proxy.py", line 363, in _callSlot
    self._raise_on_death(inner()), timeout, wait)
  File "/scratch/xctrl/karabo/extern/lib/python3.6/site-packages/karabo/middlelayer_api/eventloop.py", line 505, in sync
    return self.task.result()
  File "/scratch/xctrl/karabo/extern/lib/python3.6/site-packages/karabo/middlelayer_api/proxy.py", line 443, in _raise_on_death
    return (yield from task)
  File "/scratch/xctrl/karabo/extern/lib/python3.6/site-packages/karabo/middlelayer_api/proxy.py", line 361, in inner
    descriptor.longkey))
  File "/scratch/xctrl/karabo/extern/lib/python3.6/site-packages/karabo/middlelayer_api/signalslot.py", line 338, in call
    return (yield from self._ss.request(device, target, *args))
  File "/scratch/xctrl/karabo/extern/lib/python3.6/site-packages/karabo/middlelayer_api/eventloop.py", line 145, in request
    return (yield from future)
karabo.native.exceptions.KaraboError: 1. Exception =====>  {
    Exception Type....:  Logic Exception
    Message...........:  Command "move" is not allowed in current state "ERROR" of device "MOV_TEST/MOTOR/SERVO_1".
    File..............:  /[...]/src/karabo/core/Device.hh
    Function..........:  void karabo::core::Device<T>::ensureSlotIsValidUnderCurrentState(const string&) [with FSM = karabo::core::NoFsm; std::string = std::basic_string<char>]
    Line Number.......:  1795
    Timestamp.........:  2020-Aug-11 17:25:47.621487
}

    Timestamp.........:  2020-Aug-11 17:25:47.649369
}
"""

PYTHON_SLOT_ERROR_MESSAGE = """
Request to execute 'hello' on device 'Macro-PROBLEMATIC_SLOT-98984c38-9428-4054-9f6d-cc428c6e6e13-ProblematicSlot' failed, details:
1. Exception =====>  {
    Exception Type....:  Remote Exception from Macro-PROBLEMATIC_SLOT-98984c38-9428-4054-9f6d-cc428c6e6e13-ProblematicSlot
    Message...........:  Traceback (most recent call last):
  File "/home/carinanc/karabo/extern/lib/python3.6/site-packages/karabo/native/data/hash.py", line 591, in wrapper
    ret = yield from coro
  File "/home/carinanc/karabo/extern/lib/python3.6/site-packages/karabo/middlelayer_api/eventloop.py", line 644, in run_coroutine_or_thread
    return (yield from future)
  File "/home/carinanc/karabo/extern/lib/python3.6/site-packages/karabo/middlelayer_api/eventloop.py", line 630, in thread
    ret = f(*args, **kwargs)
  File "/home/carinanc/karabo/extern/lib/python3.6/site-packages/karabo/native/data/hash.py", line 559, in wrapper
    return self.method(device)
  File "/home/carinanc/karabo/extern/lib/python3.6/site-packages/karabo/middlelayer_api/macro.py", line 82, in wrapper
    return themethod(device)
  File "PROBLEMATIC_SLOT", line 12, in hello
  File "/home/carinanc/karabo/extern/lib/python3.6/site-packages/karabo/native/data/hash.py", line 559, in wrapper
    return self.method(device)
  File "/home/carinanc/karabo/extern/lib/python3.6/site-packages/karabo/middlelayer_api/macro.py", line 82, in wrapper
    return themethod(device)
  File "PROBLEMATIC_SLOT", line 8, in execute
RuntimeError: Problematic execute: ProblematicSlot

    Timestamp.........:  2020-Sep-07 13:28:20.319395
}
"""

PYTHON_MULTIPLE_ERROR_MESSAGE = """
Request to execute 'hello' on device 'Macro-PROBLEMATIC_SLOT-98984c38-9428-4054-9f6d-cc428c6e6e13-ProblematicSlot' failed, details:
1. Exception =====>  {
    Exception Type....:  Remote Exception from Macro-PROBLEMATIC_SLOT-98984c38-9428-4054-9f6d-cc428c6e6e13-ProblematicSlot
    Message...........:  Traceback (most recent call last):
  File "PROBLEMATIC_SLOT", line 13, in hello
  File "/home/carinanc/karabo/extern/lib/python3.6/site-packages/karabo/native/data/hash.py", line 559, in wrapper
    return self.method(device)
  File "/home/carinanc/karabo/extern/lib/python3.6/site-packages/karabo/middlelayer_api/macro.py", line 82, in wrapper
    return themethod(device)
  File "PROBLEMATIC_SLOT", line 8, in execute
NameError: name 'RunTimeError' is not defined

During handling of the above exception, another exception occurred:

Traceback (most recent call last):
  File "/home/carinanc/karabo/extern/lib/python3.6/site-packages/karabo/native/data/hash.py", line 591, in wrapper
    ret = yield from coro
  File "/home/carinanc/karabo/extern/lib/python3.6/site-packages/karabo/middlelayer_api/eventloop.py", line 644, in run_coroutine_or_thread
    return (yield from future)
  File "/home/carinanc/karabo/extern/lib/python3.6/site-packages/karabo/middlelayer_api/eventloop.py", line 630, in thread
    ret = f(*args, **kwargs)
  File "/home/carinanc/karabo/extern/lib/python3.6/site-packages/karabo/native/data/hash.py", line 559, in wrapper
    return self.method(device)
  File "/home/carinanc/karabo/extern/lib/python3.6/site-packages/karabo/middlelayer_api/macro.py", line 82, in wrapper
    return themethod(device)
  File "PROBLEMATIC_SLOT", line 15, in hello
RuntimeError: Something went wrong.

    Timestamp.........:  2020-Sep-08 00:12:24.605280
}
"""

SIMPLE_ERROR_MESSAGE = """\
Failure on request to execute 'reset' on device 'plcMonitor. Request not answered within 5 seconds."""

UNKNOWN_ERROR_MESSAGE = 42 * "karabo"

KARABO_ERROR_MESSAGE = """
Failure on request to execute 'faultySlot' on device 'XHQ_EG_DG/DATA/PROPERTY_TEST_MDL', details:
1. Exception =====>  {
    Exception Type....:  Remote Exception from XHQ_EG_DG/DATA/PROPERTY_TEST_MDL
    Message...........:  Traceback (most recent call last):
  File "/home/degon/Framework/src/pythonKarabo/karabo/native/schema/descriptors.py", line 657, in wrapper
    ret = await coro
  File "/home/degon/Framework/src/pythonKarabo/karabo/middlelayer_api/eventloop.py", line 1234, in run_coroutine_or_thread
    return (await future)
  File "/home/degon/Framework/src/pythonKarabo/karabo/middlelayer_api/eventloop.py", line 1207, in thread
    ret = f(*args, **kwargs)
  File "/home/degon/Framework/src/pythonKarabo/karabo/native/schema/descriptors.py", line 628, in wrapper
    return self.method(device)
  File "/home/degon/Framework/src/pythonKarabo/karabo/middlelayer_devices/property_test.py", line 614, in faultySlot
    raise exception("Fauly Slot cannot be executed")
karabo.native.exceptions.KaraboError: Fauly Slot cannot be executed

    Timestamp.........:  2021-Mar-31 13:35:28.973438
}
"""

KARABO_ERROR_MESSAGE_WITH_TAGGY_TEXT = """
Failure on request to execute 'loadFromTime' on device 'Princess', details:
1. Exception =====>  {
    Exception Type....:  Remote Exception from Princess
    Message...........:  Traceback (most recent call last):
  File "/path/to/work/Framework/src/pythonKarabo/karabo/native/schema/descriptors.py", line 669, in wrapper
    ret = await coro
  File "/path/to/work/Framework/src/pythonKarabo/karabo/middlelayer_api/eventloop.py", line 400, in run_coroutine_or_thread
    return (await f(*args, **kwargs))
  File "/path/to/work/Framework/src/pythonKarabo/karabo/native/schema/descriptors.py", line 636, in wrapper
    return await self.method(device)
  File "/path/to/work/Framework/src/pythonKarabo/karabo/native/schema/basetypes.py", line 44, in wrap
    raise TypeError('cannot wrap "{}" into Karabo type'.format(type(data)))
TypeError: cannot wrap "<class 'numpy.bool_'>" into Karabo type
    Timestamp.........:  2021-Dec-02 14:29:59.046236
}
"""

EXPECTED_MESSAGE = {
    CPP_SLOT_ERROR_MESSAGE: 'Command &quot;move&quot; is not allowed in current state &quot;ERROR&quot; of device &quot;MOV_TEST/MOTOR/SERVO_1&quot;.',
    PYTHON_SLOT_ERROR_MESSAGE: 'RuntimeError: Problematic execute: ProblematicSlot',
    PYTHON_MULTIPLE_ERROR_MESSAGE: 'RuntimeError: Something went wrong.',
    SIMPLE_ERROR_MESSAGE: SIMPLE_ERROR_MESSAGE,
    KARABO_ERROR_MESSAGE: "Fauly Slot cannot be executed",
    KARABO_ERROR_MESSAGE_WITH_TAGGY_TEXT: 'TypeError: cannot wrap &quot;&lt;class &#x27;numpy.bool_&#x27;&gt;&quot; into Karabo type',
    UNKNOWN_ERROR_MESSAGE: "Unknown exception",
}


def test_get_error_message():
    _assert_error_message(CPP_SLOT_ERROR_MESSAGE)
    _assert_error_message(PYTHON_SLOT_ERROR_MESSAGE)
    _assert_error_message(PYTHON_MULTIPLE_ERROR_MESSAGE)
    _assert_error_message(SIMPLE_ERROR_MESSAGE)
    _assert_error_message(KARABO_ERROR_MESSAGE)
    _assert_error_message(KARABO_ERROR_MESSAGE_WITH_TAGGY_TEXT)
    _assert_error_message(UNKNOWN_ERROR_MESSAGE)


def _assert_error_message(message):
    assert get_error_message(message) == EXPECTED_MESSAGE[message]


def test_realign_topo_hash():
    h = Hash()

    h['server.server1'] = None
    h['server.server1', ...] = {
        'host': 'BIG_IRON',
        'deviceClasses': ['FooClass', 'BarClass'],
    }
    h['server.server2'] = None
    h['server.server2', ...] = {  # None host
        'deviceClasses': ['FooClass', 'BarClass'],
    }
    h['server.server3'] = None
    h['server.server3', ...] = {
        'host': 'AN_IRON',
        'deviceClasses': ['FooClass', 'BarClass'],
    }

    servers = realign_topo_hash(h["server"], "host")

    n = iter(servers)
    item = next(n)
    assert item == "server2"
    item = next(n)
    assert item == "server3"
    item = next(n)
    assert item == "server1"


def test_encryption():
    pw = "karabo"
    cipher = "secret"
    new = encrypt(pw, cipher)
    assert new != pw
    new = decrypt(new, cipher)
    assert new == pw



#############################################################################
# Author: <kerstin.weger@xfel.eu>
# Created on February 21, 2012
# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
#############################################################################


class NavigationItemTypes:
    UNDEFINED = -1
    HOST = 0
    SERVER = 1
    CLASS = 2
    DEVICE = 3


class ConfiguratorItemType:
    LEAF = "Leaf"
    NODE = "Node"


class ProjectItemTypes:
    UNDEFINED = "Undefined"
    PROJECT = "Project"
    PROJECT_GROUP = "ProjectGroup"
    MACRO = "Macro"
    MACRO_INSTANCE = "MacroInstance"
    SCENE = "Scene"
    SERVER = "Server"
    DEVICE = "Device"
    CONFIGURATION = "Configuration"



# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
import pytest

from karabogui.binding.api import (
    VectorBoolBinding, VectorDoubleBinding, VectorInt64Binding,
    VectorStringBinding)
from karabogui.controllers.table.api import string2list
from karabogui.dialogs.api import ListEditDialog


@pytest.fixture
def dialogs(gui_app):
    string_binding = VectorStringBinding()
    string_dialog = ListEditDialog(binding=string_binding)
    string_dialog.set_list(["1", "2"])

    bool_binding = VectorBoolBinding()
    bool_dialog = ListEditDialog(binding=bool_binding)
    bool_dialog.set_list([True, False, True])

    int_binding = VectorInt64Binding()
    int_dialog = ListEditDialog(binding=int_binding)
    int_dialog.set_list([1, 2, 3])

    double_binding = VectorDoubleBinding()
    double_dialog = ListEditDialog(binding=double_binding)
    double_dialog.set_list([1.0, 2.0, 3.0])

    return string_dialog, bool_dialog, int_dialog, double_dialog


def test_allowed_choices(dialogs):

    string_dialog, bool_dialog, int_dialog, double_dialog = dialogs

    assert string_dialog._allowed_choices == {}
    assert int_dialog._allowed_choices == {}
    assert bool_dialog._allowed_choices == {"False": False, "True": True}
    assert double_dialog._allowed_choices == {}

    for dialog in dialogs:
        assert dialog.windowTitle() == "Edit list"


def test_set_list(dialogs):
    string_dialog, bool_dialog, int_dialog, double_dialog = dialogs

    assert string_dialog.list_widget.count() == 2
    string_dialog.set_list(["1", "2", "3"])
    assert string_dialog.list_widget.count() == 3
    assert string_dialog.values == ["1", "2", "3"]

    assert bool_dialog.list_widget.count() == 3
    assert bool_dialog.values == [True, False, True]
    bool_dialog.set_list([])
    assert bool_dialog.list_widget.count() == 0
    assert bool_dialog.values == []

    assert int_dialog.list_widget.count() == 3
    int_dialog.set_list([1, 2, 3, 4, 5])
    assert int_dialog.list_widget.count() == 5
    assert int_dialog.values == [1, 2, 3, 4, 5]

    assert double_dialog.list_widget.count() == 3
    double_dialog.set_list([1.1, 2.2, 3.3, 4.4, 5.5])
    assert double_dialog.list_widget.count() == 5
    assert double_dialog.values == [1.1, 2.2, 3.3, 4.4, 5.5]


def test_on_add_slot(dialogs, mocker):
    """Test the add slot of the list edit dialog"""
    string_dialog, bool_dialog, int_dialog, double_dialog = dialogs

    mock_input_dialog = mocker.patch("karabogui.dialogs.listedit.QInputDialog")

    mock_input_dialog.getText.return_value = "new", True
    string_dialog._add_clicked()
    assert string_dialog.list_widget.count() == 3
    assert string_dialog.values == ["1", "2", "new"]

    mock_input_dialog.getItem.return_value = "False", True
    bool_dialog._add_clicked()
    assert bool_dialog.list_widget.count() == 4
    assert bool_dialog.values == [True, False, True, False]

    mock_input_dialog.getInt.return_value = 55, True
    int_dialog._add_clicked()
    assert int_dialog.list_widget.count() == 4
    assert int_dialog.values == [1, 2, 3, 55]

    mock_input_dialog.getDouble.return_value = 42.2, True
    double_dialog._add_clicked()
    assert double_dialog.list_widget.count() == 4
    assert double_dialog.values == [1.0, 2.0, 3.0, 42.2]


def test_on_remove_slot(dialogs):
    """Test the remove slot of the list edit dialog"""
    for dialog in dialogs:
        expected_values = dialog.values[1:]
        dialog.list_widget.setCurrentRow(0)
        dialog._remove_clicked()
        assert dialog.list_widget.count() == len(expected_values)
        assert dialog.values == expected_values


def test_move_up_slot(dialogs):
    """Move up button on second element should move it up by one row"""
    for dialog in dialogs:
        values = dialog.values
        values[1], values[0] = values[0], values[1]
        dialog.list_widget.setCurrentRow(1)
        dialog._move_up_clicked()
        assert values == dialog.values


def test_move_down_slot(dialogs):
    """Move down button on first element should move it down by one row"""
    for dialog in dialogs:
        values = dialog.values
        values[0], values[1] = values[1], values[0]
        dialog.list_widget.setCurrentRow(0)
        dialog._move_down_clicked()
        assert values == dialog.values


def test_edit_slot(dialogs, mocker):
    string_dialog, bool_dialog, int_dialog, double_dialog = dialogs

    mock_input_dialog = mocker.patch("karabogui.dialogs.listedit.QInputDialog")

    string_dialog.list_widget.setCurrentRow(0)
    mock_input_dialog.getText.return_value = "new", True
    string_dialog._edit_clicked()
    assert string_dialog.values == ["new", "2"]

    bool_dialog.list_widget.setCurrentRow(0)
    # Item dialog with keys of choices, we return first choice
    mock_input_dialog.getItem.return_value = "False", True
    bool_dialog._edit_clicked()
    assert bool_dialog.values == [False, False, True]

    int_dialog.list_widget.setCurrentRow(0)
    mock_input_dialog.getInt.return_value = 55, True
    int_dialog._edit_clicked()
    assert int_dialog.values == [55, 2, 3]

    double_dialog.list_widget.setCurrentRow(0)
    mock_input_dialog.getDouble.return_value = 42.2, True
    double_dialog._edit_clicked()
    assert double_dialog.values == [42.2, 2.0, 3.0]


def test_binding_validation(dialogs):
    for dialog in dialogs:
        binding = dialog.binding
        values = string2list(dialog.string_values)
        assert binding.validate_trait("value", values) is not None, \
            f"Validation of {values} for {binding} failed"



# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
from karabogui.graph.common.enums import MouseTool, ROITool
from karabogui.testing import GuiTestCase

from ..toolsets import MouseModeToolset, ROIToolset


class TestMouseModeToolset(GuiTestCase):

    def test_basics(self):
        # Set up test. We create a mock slot to check the changes.
        tool = None

        def set_tool(value):
            nonlocal tool
            tool = value

        # Setup default toolset
        toolset = MouseModeToolset()
        toolset.on_trait_change(set_tool, "current_tool")

        # Check if mouse pointer is checked by default
        assert toolset.current_tool == MouseTool.Pointer
        assert toolset.buttons[MouseTool.Pointer].isChecked()

        # Click pointer for the first time
        toolset.buttons[MouseTool.Pointer].click()
        assert tool is None  # no changes
        assert toolset.buttons[MouseTool.Pointer].isChecked()

        # Click zoom for the first time, see if it's checked and
        # the pointer is unchecked
        toolset.buttons[MouseTool.Zoom].click()
        assert tool is MouseTool.Zoom
        assert toolset.buttons[MouseTool.Zoom].isChecked()
        assert not toolset.buttons[MouseTool.Pointer].isChecked()

        # Click zoom for again, see if it's unchecked and
        # the pointer is checked as default
        toolset.buttons[MouseTool.Zoom].click()
        assert tool == MouseTool.Pointer
        assert not toolset.buttons[MouseTool.Zoom].isChecked()
        assert toolset.buttons[MouseTool.Pointer].isChecked()

    def test_init_tool(self):
        # Setup toolset with desired tools
        tools = [MouseTool.Picker, MouseTool.Zoom]
        toolset = MouseModeToolset(tools=tools)

        assert len(toolset.buttons) == 2
        assert set(toolset.buttons) == set(tools)

        # Check toolset with invalid tool
        tools = [MouseTool.Picker, 'foo']
        toolset = MouseModeToolset(tools=tools)

        assert len(toolset.buttons) == 1
        assert MouseTool.Picker in toolset.buttons
        assert 'foo' not in toolset.buttons

    def test_add_tool(self):
        # Setup default toolset
        toolset = MouseModeToolset()

        # Add a valid tool
        tool = MouseTool.Picker
        toolset.add(tool)
        assert tool in toolset.buttons
        assert len(toolset.buttons) == 4

        # Add an invalid tool
        tool = "foo"
        toolset.add(tool)
        assert tool not in toolset.buttons
        assert len(toolset.buttons) == 4  # no changes


class TestROIToolset(GuiTestCase):

    def test_basics(self):
        # Setup default toolset
        toolset = ROIToolset()

        # Check if all are unchecked
        for button in toolset.buttons.values():
            assert not button.isChecked()
            actions = button.menu().actions()
            assert len(actions) == 2
            for action in actions:
                assert not action.isChecked()

    def test_click(self):
        # Set up test. We create a mock slot to check the changes.
        tool = None

        def set_tool(value):
            nonlocal tool
            tool = value

        # Setup default toolset
        toolset = ROIToolset()
        toolset.on_trait_change(set_tool, "current_tool")

        button = toolset.buttons[ROITool.Crosshair]
        button.click()
        assert tool == ROITool.Crosshair
        assert button.isChecked()

        # Mock switch actions
        actions = button.menu().actions()
        button.setDefaultAction(actions[1])
        button.click()
        assert tool == ROITool.DrawCrosshair
        assert button.isChecked()

        # Check deselect
        button.click()
        assert tool == ROITool.NoROI
        assert not button.isChecked()

    def test_check(self):
        # Set up test. We create a mock slot to check the changes.
        tool = None

        def set_tool(value):
            nonlocal tool
            tool = value

        # Setup default toolset
        toolset = ROIToolset()
        toolset.on_trait_change(set_tool, "current_tool")

        # Check the ROI.DrawRect action
        toolset.check(ROITool.DrawRect)
        button = toolset.buttons[ROITool.Rect]
        assert button.isChecked()
        assert button.defaultAction().data() == ROITool.DrawRect
        assert tool is None  # no event triggers



# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
from unittest import main, mock

from qtpy.QtCore import QPoint

from karabogui.logger import StatusLogWidget, get_logger
from karabogui.testing import GuiTestCase


class TestCase(GuiTestCase):

    def test_log_widget(self):
        logger = get_logger()
        widget = StatusLogWidget()
        log_widget = widget.log_widget
        # add handler to logger
        assert log_widget.toPlainText() == ""
        msg = "Once upon a time ..."
        logger.info(msg)
        assert msg in log_widget.toPlainText()
        assert "INFO" in log_widget.toPlainText()
        assert "ERROR" not in log_widget.toPlainText()
        logger.error(msg)
        assert msg in log_widget.toPlainText()
        assert "ERROR" in log_widget.toPlainText()
        assert "INFO" in log_widget.toPlainText()
        assert "DEBUG" not in log_widget.toPlainText()
        assert log_widget.toPlainText().count('\n') == 1
        log_widget.clear()
        assert log_widget.toPlainText().count('\n') == 0
        logger.debug(msg)
        assert "DEBUG" not in log_widget.toPlainText()

        with mock.patch("karabogui.logger.QMenu") as m:
            widget._show_context_menu(QPoint(0, 0))
            self.process_qt_events()
            m.assert_called()


if __name__ == "__main__":
    main()



# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
import pytest
from qtpy.QtCore import Qt

from karabo.common.scenemodel.api import TickSliderModel
from karabogui.binding.api import build_binding
from karabogui.testing import get_class_property_proxy, set_proxy_value

from ..slider import TickSlider
from .utils import InRangeInt, LargeRange, Object, Other


@pytest.fixture
def slider_setup(gui_app):
    proxy = get_class_property_proxy(Object.getClassSchema(), 'prop')
    controller = TickSlider(proxy=proxy, model=TickSliderModel())
    controller.create(None)
    yield controller, proxy
    # teardown
    controller.destroy()
    assert controller.widget is None


def test_slider_focus_policy(slider_setup):
    controller, _ = slider_setup
    assert controller.slider.focusPolicy() == Qt.StrongFocus


def test_slider_set_value(slider_setup):
    controller, proxy = slider_setup
    set_proxy_value(proxy, 'prop', 1.0)
    assert controller.slider.value() == 1.0
    assert controller.label.text() == "1.0"
    set_proxy_value(proxy, 'prop', 1.3)
    assert controller.slider.value() == 1.0
    assert controller.label.text() == "1.3"
    set_proxy_value(proxy, 'prop', 1.9)
    assert controller.slider.value() == 1.0
    assert controller.label.text() == "1.9"


def test_slider_edit_value(slider_setup):
    controller, proxy = slider_setup
    controller.slider.valueChanged.emit(3)
    assert proxy.edit_value == 3
    assert controller.label.text() == "3.0"


def test_slider_schema_update(gui_app):
    proxy = get_class_property_proxy(Other.getClassSchema(), 'prop')
    controller = TickSlider(proxy=proxy)
    controller.create(None)

    assert controller.slider.minimum() == 1
    assert controller.slider.maximum() == 4

    build_binding(Object.getClassSchema(),
                  existing=proxy.root_proxy.binding)

    assert controller.slider.minimum() == -2.0
    assert controller.slider.maximum() == 4.0


def test_slider_actions(slider_setup, mocker):
    controller, proxy = slider_setup
    controller = TickSlider(proxy=proxy, model=TickSliderModel())
    controller.create(None)
    action = controller.widget.actions()[0]
    assert action.text() == 'Tick Interval'

    dsym = 'karabogui.controllers.edit.slider.QInputDialog'
    QInputDialog = mocker.patch(dsym)
    QInputDialog.getInt.return_value = 20, True
    action.trigger()
    assert controller.model.ticks == 20
    assert controller.slider.tickInterval() == 20
    assert controller.slider.singleStep() == 20

    action = controller.widget.actions()[1]
    assert action.text() == 'Show value'
    assert controller.model.show_value
    action.trigger()
    assert not controller.model.show_value

    controller.destroy()


def test_slider_large_range(gui_app):
    proxy = get_class_property_proxy(LargeRange.getClassSchema(), 'prop')
    controller = TickSlider(proxy=proxy)
    controller.create(None)

    assert not controller.widget.isEnabled()
    controller.destroy()


def test_slider_large_int(gui_app):
    proxy = get_class_property_proxy(InRangeInt.getClassSchema(), 'prop')
    controller = TickSlider(proxy=proxy)
    controller.create(None)

    assert controller.widget.isEnabled()
    controller.destroy()



# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
from karabo.native import AccessLevel
from karabogui.binding.api import PropertyProxy
from karabogui.singletons.api import get_topology


def get_proxy(device_id: str, path: str) -> PropertyProxy:
    """Return a `PropertyProxy` instance for a given device and property path.
    """
    device_proxy = get_topology().get_device(device_id)
    return PropertyProxy(root_proxy=device_proxy, path=path)


def is_controller_enabled(proxy: PropertyProxy, level: AccessLevel) -> bool:
    """Check if the controller with `proxy` is enabled"""
    root_proxy = proxy.root_proxy
    state_binding = root_proxy.state_binding
    if state_binding is None:
        return False
    value = state_binding.value
    if not value:
        return False

    binding = proxy.binding
    is_allowed = binding.is_allowed(value)
    is_accessible = level >= binding.requiredAccessLevel
    return is_accessible and is_allowed



# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
from abc import abstractmethod

from qtpy.QtGui import QIcon, QKeySequence
from traits.api import ABCHasStrictTraits, Bool, Either, Instance, Int, String


class BaseSceneAction(ABCHasStrictTraits):
    """Base class for actions in a SceneView
    """
    # The icon for this action
    icon = Instance(QIcon)
    # The text label for the action
    text = String
    # The tooltip text shown when hovering over the action
    tooltip = String
    # A keyboard shortcut for the action (QKeySequence::StandardKey value)
    # or a QKeySequence
    shortcut = Either(Int, Instance(QKeySequence))
    # Whether or not this action is checkable
    checkable = Bool(False)

    @abstractmethod
    def perform(self, scene_view):
        """Perform the action on a SceneView instance.
        """

    def _shortcut_default(self):
        """The default value for the shortcut"""
        return QKeySequence.UnknownKey


class BaseSceneTool(ABCHasStrictTraits):
    """Base class for tools which within a SceneView.
    """
    # If True, this tool will be drawn whenever the scene is drawn
    visible = Bool(False)

    def draw(self, scene_view, painter):
        """The method which is responsible for drawing this tool.
        The tool for a SceneView will be drawn after everything else in
        the view has been drawn.

        This method is optional.
        """
        raise NotImplementedError

    @abstractmethod
    def mouse_down(self, scene_view, event):
        """A callback which is fired whenever the user clicks in the
        SceneView.
        """

    @abstractmethod
    def mouse_move(self, scene_view, event):
        """A callback which is fired whenever the user moves the mouse
        in the SceneView.
        """

    @abstractmethod
    def mouse_up(self, scene_view, event):
        """A callback which is fired whenever the user ends a mouse click
        in the SceneView.
        """



# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
from traits.api import Enum, Instance

from karabo.common.scenemodel.api import BaseWidgetObjectData
from karabo.native import Configurable, Int8, Int16, String, UInt32, UInt64
from karabogui.binding.api import IntBinding, StringBinding, build_binding
from karabogui.testing import GuiTestCase, flushed_registry

from ..api import (
    BaseBindingController, get_class_const_trait, get_compatible_controllers,
    get_model_controller, register_binding_controller)


class SomeObject(Configurable):
    for_display = String(options=['foo', 'bar', 'baz'])
    for_editing = String()
    unsupported = Int8()


class Integers(Configurable):
    int8 = Int8()
    int16 = Int16()
    uint32 = UInt32()
    uint64 = UInt64()


class UniqueWidgetModel(BaseWidgetObjectData):
    pass  # Satisfy the uniqueness check in register_binding_controller


def _options_checker(binding):
    return len(binding.options) > 0


def test_registry():
    with flushed_registry():
        @register_binding_controller(klassname='Editor',
                                     binding_type=StringBinding)
        class DisplayWidget(BaseBindingController):
            model = Instance(UniqueWidgetModel)

        @register_binding_controller(binding_type=StringBinding,
                                     klassname='Displayer',
                                     is_compatible=_options_checker)
        class DisplayWidgetOptions(BaseBindingController):
            model = Instance(UniqueWidgetModel)

        @register_binding_controller(klassname='Editor',
                                     binding_type=StringBinding, can_edit=True)
        class EditWidget(BaseBindingController):
            model = Instance(UniqueWidgetModel)

        schema = SomeObject.getClassSchema()
        binding = build_binding(schema)

        bind = binding.value.for_display
        for_display_widgets = get_compatible_controllers(bind)
        assert DisplayWidget in for_display_widgets
        assert DisplayWidgetOptions in for_display_widgets
        assert EditWidget not in for_display_widgets

        bind = binding.value.for_editing
        for_editing_widgets = get_compatible_controllers(bind, can_edit=True)
        assert DisplayWidget not in for_editing_widgets
        assert DisplayWidgetOptions not in for_editing_widgets
        assert EditWidget in for_editing_widgets
        for_editing_widgets = get_compatible_controllers(bind)
        assert DisplayWidget in for_editing_widgets
        assert DisplayWidgetOptions not in for_editing_widgets
        assert EditWidget not in for_editing_widgets

        bind = binding.value.unsupported
        assert get_compatible_controllers(bind) == []


def test_expanded_registry():
    with flushed_registry():
        @register_binding_controller(klassname='Displayer',
                                     binding_type=IntBinding)
        class DisplayWidget(BaseBindingController):
            model = Instance(UniqueWidgetModel)

        schema = Integers.getClassSchema()
        binding = build_binding(schema)

        for name in ('int8', 'int16', 'uint32', 'uint64'):
            bind = getattr(binding.value, name)
            assert DisplayWidget in get_compatible_controllers(bind)


def test_scene_model_registry():
    class DisplayModel(BaseWidgetObjectData):
        pass

    class EditModel(BaseWidgetObjectData):
        def _parent_component_default(self):
            return 'EditableApplyLaterComponent'

    class SharedModel(BaseWidgetObjectData):
        klass = Enum('Foo', 'Bar')

    class UnusedModel(BaseWidgetObjectData):
        pass

    with flushed_registry():
        @register_binding_controller(klassname='Displayer',
                                     binding_type=StringBinding)
        class DisplayWidget(BaseBindingController):
            model = Instance(DisplayModel)

        @register_binding_controller(klassname='Editor',
                                     binding_type=StringBinding, can_edit=True)
        class EditWidget(BaseBindingController):
            model = Instance(EditModel)

        @register_binding_controller(klassname='Bar',
                                     binding_type=StringBinding)
        class BarWidget(BaseBindingController):
            model = Instance(SharedModel)

        @register_binding_controller(klassname='Foo',
                                     binding_type=StringBinding)
        class FooWidget(BaseBindingController):
            model = Instance(SharedModel)

        controller = get_model_controller(DisplayModel())
        assert controller is DisplayWidget

        controller = get_model_controller(EditModel())
        assert controller is EditWidget

        model = SharedModel(klass='Foo')
        controller = get_model_controller(model)
        assert controller is FooWidget

        model.klass = 'Bar'
        controller = get_model_controller(model)
        assert controller is BarWidget

        model = UnusedModel()
        assert get_model_controller(model) is None


def test_display_edit_overlap():
    # IntLineEdit should be returned for both edit and non-edit widgets
    from karabo.common.scenemodel.api import IntLineEditModel

    from ..edit.lineedit import IntLineEdit

    model = IntLineEditModel(parent_component='DisplayComponent')
    assert get_model_controller(model) is IntLineEdit

    model = IntLineEditModel(
        parent_component='EditableApplyLaterComponent')
    assert get_model_controller(model) is IntLineEdit


# Use a GuiTestCase to ensure a populated controller registry
class TestEditableControllerModels(GuiTestCase):
    def test_models_are_editable(self):
        from karabogui.controllers import registry

        editable_controllers = set()
        for klasses in registry._controller_registry.values():
            for klass in klasses:
                if get_class_const_trait(klass, '_can_edit'):
                    editable_controllers.add(klass)

        for klass in editable_controllers:
            modelklass = klass.class_traits()['model'].trait_type.klass
            has_klass = 'klass' in modelklass.class_traits()
            if not has_klass:
                assert modelklass().parent_component != 'DisplayComponent'

    def test_models_are_valid(self):
        from karabogui.controllers import registry

        for klasses in registry._controller_registry.values():
            for klass in klasses:
                assert klass().model is not None



# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
import pytest
from qtpy.QtCore import (
    QEvent, QItemSelectionModel, QMimeData, QModelIndex, QPoint, Qt)
from qtpy.QtGui import QDropEvent, QKeySequence

from karabo.common.api import State
from karabo.common.scenemodel.api import (
    FilterTableElementModel, TableElementModel)
from karabo.native import (
    AccessLevel, AccessMode, Bool, Configurable, Double, Hash, Int32,
    MetricPrefix, String, Unit, VectorHash, VectorString)
from karabogui.binding.config import apply_configuration
from karabogui.controllers.table.api import (
    BaseFilterTableController, BaseTableController, KaraboTableView,
    StringButtonDelegate, TableButtonDelegate, create_mime_data, list2string)
from karabogui.testing import (
    access_level, get_property_proxy, keySequence, set_proxy_value, singletons)
from karabogui.topology.api import SystemTopology

DELEGATES = "karabogui.controllers.table.delegates"


class ButtonSchema(Configurable):
    arch = String(
        displayType="TableStringButton",
        defaultValue="deviceScene|device_id=XHQ_EG_DG/CAM/CAMERA&name=scene")
    bar = String(
        displayType="TableStringButton",
        defaultValue="url|http://www.xfel.eu")
    foo = Bool(
        displayType="TableBoolButton",
        defaultValue=False,
        displayedName="Foo")


class TableSchema(Configurable):
    arch = String(
        defaultValue="NoString")
    foo = Bool(
        defaultValue=False,
        displayedName="Foo")
    bar = String(
        enum=State,
        displayType="State",
        displayedName="Bar",
        defaultValue="ON",
        accessMode=AccessMode.READONLY)
    cat = Int32(
        defaultValue=1,
        displayedName="Cat",
        options=[1, 2, 3])
    dog = Double(
        displayedName="Dog",
        defaultValue=2.0)
    eagle = VectorString(
        displayedName="Eagle",
        description="This is a description for a table",
        unitSymbol=Unit.METER,
        metricPrefixSymbol=MetricPrefix.MILLI,
        defaultValue=[])


class Object(Configurable):
    prop = VectorHash(
        rows=TableSchema,
        accessMode=AccessMode.READONLY)


class ButtonObject(Configurable):
    prop = VectorHash(
        rows=ButtonSchema)


class StateObject(Configurable):
    state = String(
        defaultValue=State.ON)
    prop = VectorHash(
        rows=ButtonSchema,
        requiredAccessLevel=AccessLevel.OPERATOR,
        accessMode=AccessMode.READONLY,
        allowedStates=[State.ON])


def assert_model_row(controller, row, first_column, second_column,
                     third_column, fourth_column, fifth_column, sixth_column,
                     size=None):
    model = controller.sourceModel()
    # Boolean is represented as Qt value in table model

    assert model.index(row, 0).data() == first_column
    qt_bool = Qt.Checked if second_column else Qt.Unchecked
    assert model.index(row, 1).data(role=Qt.CheckStateRole) == qt_bool
    assert model.index(row, 2).data() == third_column
    # integer and floating point are string data to outside
    assert model.index(row, 3).data() == str(fourth_column)
    assert model.index(row, 4).data() == str(fifth_column)
    # Vector data is always retrieved as string list
    assert model.index(row, 5).data() == list2string(sixth_column)

    data = controller.sourceModel()._data
    # Our Hash still has real booleans, but the model does not!
    assert data[row], Hash("arch", first_column,
                           "foo", second_column,
                           "bar", third_column,
                           "cat", fourth_column,
                           "dog", fifth_column,
                           "eagle", sixth_column)
    if size is not None:
        assert data == size


@pytest.fixture
def table_model_view_setup(gui_app):
    proxy = get_property_proxy(Object.getClassSchema(), "prop")
    model = TableElementModel()
    controller = BaseTableController(proxy=proxy, model=model)
    controller.create(None)
    assert controller.isReadOnly()
    controller.set_read_only(False)
    table_hash = Hash(
        "prop",
        [Hash("arch", "a", "foo", True, "bar", "ACTIVE", "cat", 1,
              "dog", 2.0, "eagle", ["a"]),
         Hash("arch", "b", "foo", False, "bar", "ERROR", "cat", 3,
              "dog", 4.0, "eagle", ["b"]),
         Hash("arch", "c", "foo", False, "bar", "UNKNOWN", "cat", 2,
              "dog", 1.0, "eagle", ["c"]),
         Hash("arch", "d", "foo", True, "bar", "CHANGING", "cat", 1,
              "dog", 5.0, "eagle", ["d"])])
    apply_configuration(table_hash, proxy.root_proxy.binding)
    yield controller, proxy, table_hash
    # teardown
    controller.destroy()


def test_table_model_view_focus_policy(table_model_view_setup):
    controller, _, _ = table_model_view_setup
    assert controller.widget.focusPolicy() == Qt.StrongFocus


def test_table_model_view_deviceId(table_model_view_setup):
    controller, _, _ = table_model_view_setup
    deviceId = controller.getInstanceId()
    assert deviceId == "TestDevice"


def test_table_model_view_set_value(table_model_view_setup):
    controller, _, _ = table_model_view_setup
    assert_model_row(controller, 0, "a", True, "ACTIVE", 1, 2.0, ["a"])
    assert_model_row(controller, 1, "b", False, "ERROR", 3, 4.0, ["b"])
    assert_model_row(controller, 2, "c", False, "UNKNOWN", 2, 1.0, ["c"])
    assert_model_row(controller, 3, "d", True, "CHANGING", 1, 5.0, ["d"])


def test_table_model_view_header(table_model_view_setup):
    controller, _, _ = table_model_view_setup
    model = controller.sourceModel()
    header_keys = model._header
    assert header_keys == ["arch", "foo", "bar", "cat", "dog", "eagle"]
    foo = model.headerData(1, orientation=Qt.Horizontal,
                           role=Qt.DisplayRole)
    assert foo == "Foo"
    bar = model.headerData(2, orientation=Qt.Horizontal,
                           role=Qt.DisplayRole)
    assert bar == "Bar"
    cat = model.headerData(3, orientation=Qt.Horizontal,
                           role=Qt.DisplayRole)
    assert cat == "Cat"
    dog = model.headerData(4, orientation=Qt.Horizontal,
                           role=Qt.DisplayRole)
    assert dog == "Dog"
    eagle = model.headerData(5, orientation=Qt.Horizontal,
                             role=Qt.DisplayRole)
    assert eagle == "Eagle [mm]"
    is_none = model.headerData(1, orientation=Qt.Horizontal,
                               role=Qt.EditRole)
    assert is_none is None
    is_none = model.headerData(1, orientation=Qt.Vertical,
                               role=Qt.EditRole)
    assert is_none is None

    v = model.headerData(0, orientation=Qt.Vertical,
                         role=Qt.DisplayRole)
    assert v == "0"


def test_table_model_view_move_row_up(table_model_view_setup):
    controller, _, _ = table_model_view_setup
    assert_model_row(controller, 0, "a", True, "ACTIVE", 1, 2.0, ["a"])
    assert_model_row(controller, 1, "b", False, "ERROR", 3, 4.0, ["b"])

    # Move it
    model = controller.sourceModel()
    model.move_row_up(1)

    assert_model_row(controller, 0, "b", False, "ERROR", 3, 4.0, ["b"])
    assert_model_row(controller, 1, "a", True, "ACTIVE", 1, 2.0, ["a"])

    # --- Corner case ---
    model.move_row_up(0)
    assert_model_row(controller, 0, "b", False, "ERROR", 3, 4.0, ["b"])
    assert_model_row(controller, 1, "a", True, "ACTIVE", 1, 2.0, ["a"])


def test_table_model_view_move_row_down(table_model_view_setup):
    controller, _, _ = table_model_view_setup

    assert_model_row(controller, 0, "a", True, "ACTIVE", 1, 2.0, ["a"])
    assert_model_row(controller, 1, "b", False, "ERROR", 3, 4.0, ["b"])

    # Move it
    model = controller.sourceModel()
    model.move_row_down(0)

    assert_model_row(controller, 0, "b", False, "ERROR", 3, 4.0, ["b"])
    assert_model_row(controller, 1, "a", True, "ACTIVE", 1, 2.0, ["a"])
    assert_model_row(controller, 2, "c", False, "UNKNOWN", 2, 1.0, ["c"])
    assert_model_row(controller, 3, "d", True, "CHANGING", 1, 5.0, ["d"])

    # --- Corner case ---
    model.move_row_down(3)
    assert_model_row(controller, 0, "b", False, "ERROR", 3, 4.0, ["b"])
    assert_model_row(controller, 1, "a", True, "ACTIVE", 1, 2.0, ["a"])
    assert_model_row(controller, 2, "c", False, "UNKNOWN", 2, 1.0, ["c"])
    assert_model_row(controller, 3, "d", True, "CHANGING", 1, 5.0, ["d"])


def test_table_model_view_duplicate_row(table_model_view_setup):
    controller, _, _ = table_model_view_setup

    assert_model_row(controller, 0, "a", True, "ACTIVE", 1, 2.0, ["a"])
    assert_model_row(controller, 1, "b", False, "ERROR", 3, 4.0, ["b"])
    model = controller.sourceModel()
    model.duplicate_row(0)
    assert_model_row(controller, 0, "a", True, "ACTIVE", 1, 2.0, ["a"])
    assert_model_row(controller, 1, "a", True, "ACTIVE", 1, 2.0, ["a"])
    assert_model_row(controller, 2, "b", False, "ERROR", 3, 4.0, ["b"])


def test_table_model_view_empty_value(table_model_view_setup):
    controller, proxy, table_hash = table_model_view_setup
    model = controller.sourceModel()
    assert model._data != []
    model.clear_model()
    assert model._data == []
    assert model.rowCount(QModelIndex()) == 0
    # And a new round of applying configuration
    apply_configuration(table_hash, proxy.root_proxy.binding)
    assert model.rowCount(QModelIndex()) == 4
    # Set empty list
    apply_configuration(Hash("prop", []), proxy.root_proxy.binding)
    assert model.rowCount(QModelIndex()) == 0
    # Fill up again and decrease to 3
    apply_configuration(table_hash, proxy.root_proxy.binding)
    assert model.rowCount(QModelIndex()) == 4

    h = Hash(
        "prop",
        [Hash("arch", "a", "foo", True, "bar", "ACTIVE", "cat", 1,
              "dog", 2.0, "eagle", ["a"]),
         Hash("arch", "b", "foo", False, "bar", "ERROR", "cat", 3,
              "dog", 4.0, "eagle", ["b"]),
         Hash("arch", "c", "foo", False, "bar", "UNKNOWN", "cat", 2,
              "dog", 1.0, "eagle", ["c"])])
    apply_configuration(h, proxy.root_proxy.binding)
    assert model.rowCount(QModelIndex()) == 3

    # Test model data
    value = controller.get_model_data(0, 0)
    assert value == ("arch", "a")
    value = controller.get_model_data(0, 1)
    assert value == ("foo", True)
    value = controller.get_model_data(0, 2)
    assert value == ("bar", "ACTIVE")

    value = controller.get_model_data(1, 0)
    assert value == ("arch", "b")
    value = controller.getModelData(1, 0)
    assert value == ("arch", "b")


def test_table_model_view_background_value(table_model_view_setup):
    controller, _, _ = table_model_view_setup

    model = controller.sourceModel()
    index = model.index(0, 1)
    assert index.data(role=Qt.BackgroundRole) is None
    index = model.index(0, 2)
    brush = index.data(role=Qt.BackgroundRole)
    assert brush is not None
    color = brush.color().getRgb()
    assert color == (120, 255, 0, 255)
    # Set new state
    model.setData(index, value="ERROR", role=Qt.EditRole)
    brush = index.data(role=Qt.BackgroundRole)
    color = brush.color().getRgb()
    assert color == (255, 0, 0, 255)

    # Set wrong state
    model.setData(index, value=" NOSTATE", role=Qt.EditRole)
    brush = index.data(role=Qt.BackgroundRole)
    assert brush is None


def test_table_model_view_set_boolean_data(table_model_view_setup):
    controller, _, _ = table_model_view_setup
    model = controller.sourceModel()
    index = model.index(0, 1)
    assert_model_row(controller, 0, "a", True, "ACTIVE", 1, 2.0, ["a"])
    model.setData(index, value=Qt.Unchecked, role=Qt.CheckStateRole)
    assert_model_row(controller, 0, "a", False, "ACTIVE", 1, 2.0, ["a"])
    model.setData(index, value=Qt.Checked, role=Qt.CheckStateRole)
    assert_model_row(controller, 0, "a", True, "ACTIVE", 1, 2.0, ["a"])


def test_table_model_view_set_vector_data(table_model_view_setup):
    controller, _, _ = table_model_view_setup
    model = controller.sourceModel()
    assert_model_row(controller, 0, "a", True, "ACTIVE", 1, 2.0, ["a"])
    index = model.index(0, 5)
    # The model expects string data for lists
    model.setData(index, value=list2string(["a", "b", "c"]))
    data = index.data(role=Qt.DisplayRole)
    assert data == "a,b,c"
    with pytest.raises(AttributeError):
        # List does not have split attribute ...
        model.setData(index, value=["a", "b", "c"])


def test_table_model_view_tooltip_data(table_model_view_setup):
    controller, _, _ = table_model_view_setup
    model = controller.sourceModel()
    index = model.index(0, 5)
    data = index.data(role=Qt.ToolTipRole)
    assert data == "a"
    ret = model.setData(index, "Not allowed", role=Qt.ToolTipRole)
    assert not ret

    # Check the header
    data = model.headerData(section=5, orientation=Qt.Horizontal,
                            role=Qt.ToolTipRole)
    assert "key" in data
    assert "metricPrefixSymbol" in data
    assert "unitSymbol" in data
    assert "defaultValue" in data
    assert "valueType" in data
    assert "VECTOR_STRING" in data
    assert "description" in data
    assert "This is a description for a table" in data
    assert "minExc" not in data
    assert "minInc" not in data


def test_table_model_view_invalid_model_index(table_model_view_setup):
    controller, _, _ = table_model_view_setup
    model = controller.sourceModel()
    ret = model.setData(QModelIndex(), 2, Qt.EditRole)
    assert not ret
    assert model.data(QModelIndex()) is None
    assert model.flags(QModelIndex()) == 0


def test_table_model_view_model_flags(table_model_view_setup):
    controller, _, _ = table_model_view_setup
    model = controller.sourceModel()
    # 1. User checkable boolean
    index = model.index(0, 1)
    flags = index.flags()
    assert (flags & Qt.ItemIsUserCheckable) == Qt.ItemIsUserCheckable
    assert (flags & Qt.ItemIsEditable) != Qt.ItemIsEditable

    # 2. Readonly string
    index = model.index(0, 2)
    flags = index.flags()
    assert (flags & Qt.ItemIsUserCheckable) != Qt.ItemIsUserCheckable
    assert (flags & Qt.ItemIsEditable) != Qt.ItemIsEditable

    # 3. reconfigurable int
    index = model.index(0, 3)
    flags = index.flags()
    assert (flags & Qt.ItemIsUserCheckable) != Qt.ItemIsUserCheckable
    assert (flags & Qt.ItemIsEditable) == Qt.ItemIsEditable


def test_table_model_view_read_only_table(table_model_view_setup):
    controller, _, _ = table_model_view_setup
    controller.set_read_only(True)


def test_table_model_view_view_menu_select(table_model_view_setup, mocker):
    controller, _, _ = table_model_view_setup
    menu_path = "karabogui.controllers.table.controller.QMenu"
    pos = QPoint(0, 0)
    menu = mocker.patch(menu_path)
    model = controller.sourceModel()
    index = model.index(2, 0)

    view = controller.widget
    selection = view.selectionModel()
    selection.setCurrentIndex(index,
                              QItemSelectionModel.ClearAndSelect)
    # Assert the selection
    assert view.currentIndex() == index
    assert controller.currentIndex() == index

    controller._context_menu(pos)
    menu.assert_called_once()

    model.setData(index, "Yes A String")
    assert index.data() == "Yes A String"
    controller._set_index_default()
    assert index.data() == "NoString"

    # Invalid index also provides a model!
    menu.reset_mock()
    selection.setCurrentIndex(QModelIndex(),
                              QItemSelectionModel.ClearAndSelect)
    assert view.currentIndex() == QModelIndex()

    controller._context_menu(pos)
    menu.assert_called_once()


def test_table_model_view_drag_drop_view(table_model_view_setup):
    """Test the drag & drop of the table view"""
    controller, _, _ = table_model_view_setup

    items = QMimeData()
    items.setData("tableData", create_mime_data(row=1))

    view = controller.widget

    model = controller.sourceModel()
    index = model.index(0, 0)
    assert index.data() == "a"

    event = QDropEvent(QPoint(0, 0), Qt.CopyAction, items,
                       Qt.LeftButton, Qt.NoModifier, QEvent.Drop)
    # Patch the source
    event.source = lambda: controller.widget

    view.dropEvent(event)
    assert index.data() == "b"
    index = model.index(1, 0)
    assert index.data() == "a"

    assert_model_row(controller, 3, "d", True, "CHANGING", 1, 5.0, ["d"])
    assert controller.sourceModel().rowCount(None) == 4

    items.setData("tableData", create_mime_data(row=2))
    # Drag somewhere to outer regions, it will modify last row
    event = QDropEvent(QPoint(0, 1000), Qt.CopyAction, items,
                       Qt.LeftButton, Qt.NoModifier, QEvent.Drop)
    event.source = lambda: controller.widget
    view.dropEvent(event)
    assert controller.sourceModel().rowCount(None) == 4
    assert_model_row(controller, 3, "c", False, "UNKNOWN", 2, 1.0, ["c"])


def test_table_model_view_button_delegate(table_model_view_setup):
    """Test the table button delegate of the karabo table view"""
    clicked = False

    class Delegate(TableButtonDelegate):

        def get_button_text(self, index):
            return "Boolean"

        def isEnabled(self, index=None):
            return index.data(role=Qt.CheckStateRole) == Qt.Checked

        def click_action(self, index):
            nonlocal clicked
            clicked = True

    controller, _, _ = table_model_view_setup

    # XXX: Find a better of testing the delegate!
    button_delegate = Delegate()
    controller.widget.setItemDelegateForColumn(1, button_delegate)
    model = controller.sourceModel()
    idx = model.index(2, 1)
    assert not button_delegate.isEnabled(idx)
    button_delegate.click_action(idx)
    assert clicked

    idx = model.index(0, 1)
    assert button_delegate.isEnabled(idx)
    button_delegate.click_action(idx)


def test_table_model_view_string_button_delegate(gui_app, mocker):
    """Test the table string button delegate of the karabo table view"""
    s = "deviceScene|device_id=XHQ_EG_DG/CAM/CAMERA&name=scene"
    table_hash = Hash(
        "prop", [Hash("arch", s,
                      "bar", "url|https://www.xfel.eu",
                      "foo", True)])

    proxy = get_property_proxy(ButtonObject.getClassSchema(), "prop")
    model = TableElementModel()
    controller = BaseTableController(proxy=proxy, model=model)
    controller.create(None)
    apply_configuration(table_hash, proxy.root_proxy.binding)
    model = controller.sourceModel()

    # 1.1 Device scene delegate
    assert model.rowCount() == 1
    index = model.index(0, 0)
    table_view = controller.tableWidget()
    delegate = table_view.itemDelegate(index)
    assert isinstance(delegate, StringButtonDelegate)
    topology = SystemTopology()
    network = mocker.Mock()
    with singletons(topology=topology, network=network):
        delegate.click_action(index)
        # Click, but device is offline
        network.onExecuteGeneric.assert_not_called()

        device_id = proxy.root_proxy.device_id
        device_hash = Hash(device_id, "")
        device_hash[device_id, ...] = {"status": "ok",
                                       "capabilities": 0,
                                       "host": "BIG_IRON",
                                       "visibility": 4,
                                       "serverId": "swerver",
                                       "classId": "FooBar"}
        topology.initialize(Hash("device", device_hash))
        # Click again, device is online
        delegate.click_action(index)
        args = network.onExecuteGeneric.call_args
        assert "XHQ_EG_DG/CAM/CAMERA" in args[0]
        assert "requestScene" in args[0]

        # 1.2 Check the url web link delegate
        index = model.index(0, 1)
        table_view = controller.tableWidget()
        delegate = table_view.itemDelegate(index)
        assert isinstance(delegate, StringButtonDelegate)
        path = f"{DELEGATES}.webbrowser"
        web = mocker.patch(path)
        delegate.click_action(index)
        web.open_new.assert_called_with("https://www.xfel.eu")

        # Retrieve default scene if no name is provided
        index = model.index(0, 0)
        delegate = table_view.itemDelegate(index)
        s = "deviceScene|device_id=XHQ_EG_DG/CAM/CAMERA"
        table_hash = Hash(
            "prop", [Hash("arch", s,
                          "bar", "url|https://www.xfel.eu",
                          "foo", True)])
        apply_configuration(table_hash, proxy.root_proxy.binding)
        rd = mocker.patch(f"{DELEGATES}.retrieve_default_scene")
        delegate.click_action(index)
        rd.assert_called_with("XHQ_EG_DG/CAM/CAMERA")

    # teardown
    controller.destroy()


def test_table_model_view_set_data(table_model_view_setup):
    controller, proxy, _ = table_model_view_setup
    proxy.edit_value = None
    assert proxy.edit_value is None

    model = controller.sourceModel()
    index = model.index(0, 0)
    assert index.data() == "a"
    model.setData(index, "b")
    index = model.index(0, 0)
    assert index.data() == "b"
    assert proxy.edit_value is not None
    first_hash = proxy.edit_value[0]
    assert first_hash["arch"] == "b"

    index = model.index(0, 3)
    assert index.data() == "1"
    model.setData(index, "13")
    index = model.index(0, 3)
    assert index.data() == "13"
    first_hash = proxy.edit_value[0]
    assert first_hash["cat"] == 13

    index = model.index(0, 4)
    assert index.data() == "2.0"
    model.setData(index, "2.12456")
    index = model.index(0, 4)
    assert index.data() == "2.12456"
    first_hash = proxy.edit_value[0]
    assert first_hash["dog"] == 2.12456

    index = model.index(0, 5)
    assert index.data() == "a"
    model.setData(index, list2string(["a", "b", "c"]))
    index = model.index(0, 5)
    assert index.data() == "a,b,c"
    first_hash = proxy.edit_value[0]
    assert first_hash["eagle"] == ["a", "b", "c"]


def test_table_model_view_basic_settings_controller(table_model_view_setup):
    controller, _, _ = table_model_view_setup
    assert not controller.isReadOnly()
    assert isinstance(controller.widget, KaraboTableView)
    assert controller.getBindings() is not None

    # Destroy before teardown
    controller.destroy()
    assert controller.widget is None
    assert controller.sourceModel() is None


def test_table_model_view_table_actions(table_model_view_setup):
    controller, _, _ = table_model_view_setup
    actions = controller.widget.actions()
    assert len(actions) == 1
    action = actions[0]
    assert action.text() == "Resize To Contents"
    assert not action.isChecked()
    action.trigger()
    assert action.isChecked()


@pytest.fixture
def filter_table_model_view_setup(gui_app):
    proxy = get_property_proxy(Object.getClassSchema(), "prop")
    model = FilterTableElementModel()
    controller = BaseFilterTableController(proxy=proxy, model=model)
    controller.create(None)
    assert controller.isReadOnly()
    controller.set_read_only(False)
    table_hash = Hash(
        "prop",
        [Hash("arch", "a", "foo", True, "bar", "ACTIVE", "cat", 1,
              "dog", 2.0, "eagle", ["a"]),
         Hash("arch", "b", "foo", False, "bar", "ERROR", "cat", 3,
              "dog", 4.0, "eagle", ["b"]),
         Hash("arch", "c", "foo", False, "bar", "UNKNOWN", "cat", 2,
              "dog", 1.0, "eagle", ["c"]),
         Hash("arch", "d", "foo", True, "bar", "CHANGING", "cat", 1,
              "dog", 5.0, "eagle", ["d"])])
    apply_configuration(table_hash, proxy.root_proxy.binding)

    yield controller, proxy, model
    controller.destroy()


def test_filter_table_model_view_set_value(filter_table_model_view_setup):
    controller, _, _ = filter_table_model_view_setup
    assert_model_row(controller, 0, "a", True, "ACTIVE", 1, 2.0, ["a"])
    assert_model_row(controller, 1, "b", False, "ERROR", 3, 4.0, ["b"])
    assert_model_row(controller, 2, "c", False, "UNKNOWN", 2, 1.0, ["c"])
    assert_model_row(controller, 3, "d", True, "CHANGING", 1, 5.0, ["d"])


def test_filter_table_model_view_filter_combo(filter_table_model_view_setup):
    controller, _, _ = filter_table_model_view_setup
    model = controller.sourceModel()
    for i in range(controller.columnCombo.count()):
        header_text = model.headerData(i, orientation=Qt.Horizontal,
                                       role=Qt.DisplayRole)
        assert header_text == controller.columnCombo.itemText(i)
    controller.columnCombo.setCurrentIndex(2)
    controller.searchLabel.setText("ACTIVE")

    filter_model = controller.tableModel()

    # check if its in the filtered view
    index = model.index(0, 2, QModelIndex())
    # ACTIVE is visible
    assert filter_model.mapFromSource(index).isValid()
    index = model.index(1, 2, QModelIndex())
    # row with ERROR cell is not visible
    assert not filter_model.mapFromSource(index).isValid()


def test_filter_table_model_view_table_actions(filter_table_model_view_setup,
                                               mocker):
    controller, _, model = filter_table_model_view_setup
    actions = controller.widget.actions()
    assert len(actions) == 4
    assert actions[0].text() == "Resize To Contents"
    assert actions[1].text() == "Set Default Filter Column"
    assert actions[2].text() == "Sorting Enabled"
    assert actions[3].text() == "Show Filter Column Toggle"

    assert model.filterKeyColumn == 0
    path = "karabogui.controllers.table.controller.QInputDialog"
    dia = mocker.patch(path)
    dia.getInt.return_value = 2, True
    actions[1].trigger()
    assert model.filterKeyColumn == 2

    assert not controller.model.sortingEnabled
    actions[2].trigger()
    assert controller.model.sortingEnabled

    assert not controller.model.showFilterKeyColumn
    actions[3].trigger()
    assert controller.model.showFilterKeyColumn


def test_filter_table_model_view_model_data(filter_table_model_view_setup):
    controller, proxy, _ = filter_table_model_view_setup
    model = controller.tableWidget().model()

    h = Hash(
        "prop",
        [Hash("arch", "a", "foo", True, "bar", "ACTIVE", "cat", 1,
              "dog", 2.0, "eagle", ["a"]),
         Hash("arch", "b", "foo", False, "bar", "ERROR", "cat", 3,
              "dog", 4.0, "eagle", ["b"]),
         Hash("arch", "c", "foo", False, "bar", "UNKNOWN", "cat", 2,
              "dog", 1.0, "eagle", ["c"])])
    apply_configuration(h, proxy.root_proxy.binding)
    assert model.rowCount(QModelIndex()) == 3

    # Test model data
    value = controller.get_model_data(0, 0)
    assert value == ("arch", "a")
    value = controller.get_model_data(0, 1)
    assert value == ("foo", True)
    value = controller.get_model_data(0, 2)
    assert value == ("bar", "ACTIVE")
    value = controller.get_model_data(1, 0)
    assert value == ("arch", "b")


def test_filter_table_model_view_view_key_events(
        filter_table_model_view_setup):
    controller, proxy, _ = filter_table_model_view_setup

    model = controller.tableWidget().model()

    h = Hash(
        "prop",
        [Hash("arch", "a", "foo", True, "bar", "ACTIVE", "cat", 1,
              "dog", 2.0, "eagle", ["a"]),
         Hash("arch", "b", "foo", False, "bar", "ERROR", "cat", 3,
              "dog", 4.0, "eagle", ["b"]),
         Hash("arch", "c", "foo", False, "bar", "UNKNOWN", "cat", 2,
              "dog", 1.0, "eagle", ["c"])])
    apply_configuration(h, proxy.root_proxy.binding)
    assert model.rowCount(QModelIndex()) == 3

    assert not controller.isReadOnly()

    # 1. Move row up and down
    model = controller.sourceModel()
    assert model.index(0, 0).data() == "a"
    assert model.index(1, 0).data() == "b"
    controller.tableWidget().selectRow(1)
    keySequence(controller.tableWidget(), QKeySequence.SelectPreviousLine)
    assert model.index(0, 0).data() == "b"
    assert model.index(1, 0).data() == "a"
    controller.tableWidget().selectRow(0)
    keySequence(controller.tableWidget(), QKeySequence.SelectNextLine)
    assert model.index(0, 0).data() == "a"
    assert model.index(1, 0).data() == "b"

    controller.tableWidget().selectRow(3)
    # 2. Add a row
    keySequence(controller.tableWidget(), QKeySequence.New)
    assert model.rowCount(QModelIndex()) == 4

    # 3. Select last row and delete a row
    controller.tableWidget().selectRow(4)
    keySequence(controller.tableWidget(), QKeySequence.Delete)
    assert model.rowCount(QModelIndex()) == 3


def test_filter_table_model_view_table_controller_extras(gui_app):
    proxy = get_property_proxy(Object.getClassSchema(), "prop")
    model = FilterTableElementModel(sortingEnabled=True)
    controller = BaseFilterTableController(proxy=proxy,
                                           model=model)
    controller.create(None)
    # 1.1 We have sorting enabled
    assert controller.tableWidget().isSortingEnabled()
    # 1.2 Does not have any buttons
    assert not controller._table_buttons
    controller.destroy()

    proxy = get_property_proxy(ButtonObject.getClassSchema(), "prop")
    model = FilterTableElementModel(sortingEnabled=False)
    controller = BaseFilterTableController(proxy=proxy, model=model)
    controller.create(None)
    # 2.1 We have sorting enabled
    assert not controller.tableWidget().isSortingEnabled()
    # 2.2 Does not have any buttons
    assert controller._table_buttons


def test_filter_table_model_view_table_controller_state(gui_app):
    proxy = get_property_proxy(StateObject.getClassSchema(), "prop")
    model = TableElementModel()
    controller = BaseTableController(proxy=proxy, model=model)
    controller.create(None)
    controller.set_read_only(False)
    assert controller.widget.isEnabled()
    set_proxy_value(proxy, "state", "CHANGING")
    assert not controller.widget.isEnabled()
    set_proxy_value(proxy, "state", "ON")
    assert controller.widget.isEnabled()

    # We have buttons and are also disabled
    controller.set_read_only(True)
    set_proxy_value(proxy, "state", "CHANGING")
    assert not controller.widget.isEnabled()
    set_proxy_value(proxy, "state", "ON")
    assert controller.widget.isEnabled()

    with access_level(AccessLevel.OBSERVER):
        set_proxy_value(proxy, "state", "ON")
        assert not controller.widget.isEnabled()

    controller.destroy()



# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
from platform import system

import numpy as np
import pytest
from qtpy.QtWidgets import QGraphicsTextItem

from karabo.common.scenemodel.api import VectorRollGraphModel
from karabo.native import (
    Configurable, Hash, Int32, NDArray, Timestamp, VectorInt32)
from karabogui.graph.common.api import AuxPlots
from karabogui.testing import (
    get_class_property_proxy, set_proxy_hash, set_proxy_value)

from ..vector_roll_graph import ArrayRollGraph


class Object(Configurable):
    prop = VectorInt32(defaultValue=[1, 2, 3])


class NDArrayObject(Configurable):
    prop = NDArray(
        defaultValue=np.arange(10, dtype=np.int32),
        shape=(10,),
        dtype=Int32)


@pytest.fixture
def vector_roll_graph_setup(gui_app, mocker):
    schema = Object.getClassSchema()
    proxy = get_class_property_proxy(schema, "prop")
    controller = ArrayRollGraph(proxy=proxy, model=VectorRollGraphModel())

    mocker.patch.object(QGraphicsTextItem, "setHtml")
    controller.create(None)
    assert controller.widget is not None

    yield controller, proxy

    controller.destroy()
    assert controller.widget is None


def test_vector_roll_graph_configuration(vector_roll_graph_setup):
    """Assert that the vector roll auxiliar plots do not smooth the
    images"""
    controller, _ = vector_roll_graph_setup
    aux_plots = controller.widget._aux_plots
    controller = aux_plots._aggregators[AuxPlots.ProfilePlot]
    assert not controller.smooth


@pytest.mark.skipif(system() == "Windows",
                    reason="image.data is None in Windows tests")
def test_vector_roll_graph_set_value(vector_roll_graph_setup):
    """Test the value setting in VectorRollGraph"""
    controller, proxy = vector_roll_graph_setup
    plot = controller._plot
    assert plot is not None
    value = [2, 4, 6]
    set_proxy_value(proxy, "prop", value)
    image = controller._image
    np.testing.assert_almost_equal(image.data[0], value)


@pytest.mark.skipif(system() == "Windows",
                    reason="image.data is None in Windows tests")
def test_vector_roll_graph_set_value_timestamp(vector_roll_graph_setup):
    """Test the value setting with same timestamp in VectorRollGraph"""
    controller, proxy = vector_roll_graph_setup
    plot = controller._plot
    assert plot is not None
    timestamp = Timestamp()
    h = Hash("prop", [2, 4, 6])
    set_proxy_hash(proxy, h, timestamp)
    image = controller._image
    np.testing.assert_almost_equal(image.data[0], [2, 4, 6])
    h = Hash("prop", [22, 34, 16])
    set_proxy_hash(proxy, h, timestamp)
    image = controller._image
    np.testing.assert_almost_equal(image.data[0], [2, 4, 6])


def test_vector_roll_graph_image_stack_configuration(vector_roll_graph_setup,
                                                     mocker):
    _, proxy = vector_roll_graph_setup
    controller = ArrayRollGraph(proxy=proxy, model=VectorRollGraphModel())
    mocker.patch.object(QGraphicsTextItem, "setHtml")
    controller.create(None)

    action = controller.widget.actions()[4]
    assert action.text() == "Image Size"

    dsym = "karabogui.controllers.display.vector_roll_graph.QInputDialog"
    QInputDialog = mocker.patch(dsym)
    QInputDialog.getInt.return_value = 20, True
    action.trigger()
    assert controller.model.maxlen == 20
    assert controller._image.stack == 20
    assert controller._image.data is None

    controller.destroy()


@pytest.mark.skipif(system() == "Windows",
                    reason="image.data is None in Windows tests")
def test_set_value(gui_app):
    # setup
    schema = NDArrayObject.getClassSchema()
    proxy = get_class_property_proxy(schema, "prop")
    controller = ArrayRollGraph(proxy=proxy)
    controller.create(None)
    assert controller.widget is not None

    # test body
    value = np.array(
        [2, 3, 2, 3, 2, 3, 2, 3, 2, 3],
        dtype=np.int32)
    array_hash = Hash("type", 12,
                      "data", value.tobytes())
    set_proxy_value(proxy, "prop", array_hash)

    image = controller._image
    np.testing.assert_almost_equal(image.data[0], value)

    # teardown
    controller.destroy()
    assert controller.widget is None



# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
import datetime
from collections import OrderedDict
from itertools import cycle
from weakref import WeakValueDictionary

from qtpy import uic
from qtpy.QtCore import QDateTime, QTimer
from qtpy.QtGui import QColor
from qtpy.QtWidgets import QAction, QDialog, QVBoxLayout, QWidget
from traits.api import Bool, Dict, Instance, Int, Set, String, WeakRef

from karabo.common.scenemodel.api import (
    KARABO_CURVE_OPTIONS, AlarmGraphModel, CurveType, StateGraphModel,
    TrendGraphModel, build_graph_config, extract_graph_curve_option,
    restore_graph_config)
from karabogui import icons
from karabogui.binding.api import (
    BoolBinding, FloatBinding, IntBinding, StringBinding)
from karabogui.controllers.api import (
    HIDDEN, ONE_DAY, ONE_HOUR, ONE_WEEK, TEN_MINUTES, UPTIME,
    BaseBindingController, Curve, get_start_end_date_time,
    register_binding_controller, with_display_type)
from karabogui.dialogs.api import RequestTimeDialog
from karabogui.graph.common.api import AxisType, create_button, get_pen_cycler
from karabogui.graph.common.const import ALARM_INTEGER_MAP, STATE_INTEGER_MAP
from karabogui.graph.plots.base import KaraboPlotView
from karabogui.graph.plots.dialogs import CurveOptionsDialog
from karabogui.graph.plots.utils import create_curve_options

from .util import get_ui_file

# NOTE: We limit ourselves to selected karabo actions!
BASE_ACTIONS = ["x_grid", "y_grid", "y_invert", "y_log", "axes", "view"]
TIME_ACTIONS = BASE_ACTIONS + ["y_range"]

MIN_TIMESTAMP = datetime.datetime(1970, 1, 31).timestamp()
MAX_TIMESTAMP = datetime.datetime(2038, 12, 31).timestamp()


def curve_to_actions(axis_type):
    """Return the allowed actions for the curve type"""
    return TIME_ACTIONS if axis_type is AxisType.Time else BASE_ACTIONS


def curve_to_axis(value):
    """Return the axis type of the current curve configuration"""
    return [AxisType.Time, AxisType.State, AxisType.Alarm][value]


class BaseSeriesGraph(BaseBindingController):
    """The BaseClass for all trendline controllers"""
    _plot = WeakRef(KaraboPlotView)
    _timer = Instance(QTimer)
    _start_time = Instance(QDateTime)

    # the plots from pyqtgraph
    _curves = Instance(WeakValueDictionary, args=())

    # Holds if the user is rescaling via mouse interaction
    _button_scale = Bool(False)
    _curve_points = Dict
    _curves_start = Set
    _pens = Instance(cycle, factory=get_pen_cycler, args=())

    # Map time strs to QPushButtons
    _x_axis_buttons = Instance(OrderedDict, args=())
    _x_detail = String(UPTIME)

    # Our type of curve
    curve_type = Int(0)

    _window_action = Instance(QAction)
    _window = Int(1)

    def create_widget(self, parent):
        """Setup all widgets correctly"""
        widget = QWidget(parent)
        uic.loadUi(get_ui_file("ui_trendline.ui"), widget)
        self._start_time = QDateTime.currentDateTime()

        widget.dt_start.setDateTime(self._start_time)
        widget.dt_end.setDateTime(self._start_time)

        # Init x-axis buttons
        self._x_axis_buttons[widget.bt_one_week] = ONE_WEEK
        self._x_axis_buttons[widget.bt_one_day] = ONE_DAY
        self._x_axis_buttons[widget.bt_one_hour] = ONE_HOUR
        self._x_axis_buttons[widget.bt_ten_minutes] = TEN_MINUTES
        self._x_axis_buttons[widget.bt_window] = HIDDEN
        self._x_axis_buttons[widget.bt_uptime] = UPTIME
        # Hide the hidden button for window mode
        widget.bt_window.setVisible(False)
        widget.bg_x_axis.buttonClicked.connect(self._x_axis_btns_toggled)

        axis_type = curve_to_axis(self.curve_type)
        actions = curve_to_actions(axis_type)
        self._plot = KaraboPlotView(axis=axis_type,
                                    actions=actions,
                                    parent=widget.time_frame)
        self._plot.stateChanged.connect(self._change_model)
        self._plot.add_legend(visible=False)
        self._plot.add_cross_target()
        toolbar = self._plot.add_toolbar()
        self._plot.enable_export()
        self._plot.enable_data_toggle(activate=True)

        _btn_reset = create_button(
            checkable=False,
            icon=icons.reset,
            tooltip="Reload the view and request historic data",
            on_clicked=self._purge_curves)
        toolbar.add_button(_btn_reset)

        # Restore previous configuration!
        self._plot.restore(build_graph_config(self.model))

        viewbox = self._plot.plotItem.vb
        viewbox.autorange_enabled = False
        viewbox.sigRangeChangedManually.connect(self._range_change_manually)
        viewbox.middleButtonClicked.connect(self._autorange_requested)

        dialog_ac = QAction(icons.clock, "Request Time", parent=widget)
        dialog_ac.triggered.connect(self._request_dialog)
        viewbox.add_action(dialog_ac, separator=False)

        window_ac = QAction("Window Mode", parent=widget)
        window_ac.triggered.connect(widget.bt_window.click)
        window_ac.setCheckable(True)
        viewbox.add_action(window_ac, separator=False)
        self._window_action = window_ac

        # Update datetime widgets everytime range changes and limit zoom
        self._plot.plotItem.setLimits(xMin=MIN_TIMESTAMP, xMax=MAX_TIMESTAMP)
        self._plot.plotItem.sigXRangeChanged.connect(self._update_date_widgets)

        layout = QVBoxLayout()
        layout.setSizeConstraint(QVBoxLayout.SetNoConstraint)
        layout.setContentsMargins(0, 0, 0, 0)
        layout.addWidget(self._plot)
        widget.time_frame.setLayout(layout)

        # XXX: Have a 1s timeout to request data, thus avoid frequent
        # re-loading while scaling
        self._timer = QTimer(parent)
        self._timer.setInterval(1000)
        self._timer.setSingleShot(True)
        self._timer.timeout.connect(self._update_intervals)

        # Add the first curve
        self.add_proxy(self.proxy)

        # Add the plot actions to the general widget
        widget.addActions(self._plot.actions())

        # Finally, add the plot options
        curve_options = QAction("Curve Options", widget)
        curve_options.setIcon(icons.settings)
        curve_options.triggered.connect(self.show_options_dialog)
        widget.addAction(curve_options)

        return widget

    def add_proxy(self, proxy):
        if proxy in self._curve_points:
            return

        name = proxy.key
        pen = next(self._pens)
        options = extract_graph_curve_option(self.model, name)
        if options:
            pen_color = options["pen_color"]
            pen.setColor(QColor(pen_color))
            name = options["name"]

        curve = self._plot.add_curve_item(
            name=name, pen=pen, connect="all")
        curve.sigPlotChanged.connect(self._update_x_range)

        self._curves[proxy] = curve
        self._curve_points[proxy] = Curve(
            proxy=proxy, curve=curve,
            curve_type=self.curve_type)
        self._curves_start.add(proxy)

        if len(self._curves) > 1:
            self._plot.set_legend(True)

        return True

    def remove_proxy(self, proxy):
        """Remove an additional proxy curve from the controller"""
        options = [option for option in self.model.curve_options
                   if option.key != proxy.key]
        self.model.trait_set(curve_options=options)

        self._curves.pop(proxy)
        self._curves_start.discard(proxy)
        curve = self._curve_points.pop(proxy)
        item = curve.curve
        self._plot.remove_item(item)
        item.sigPlotChanged.disconnect()
        item.deleteLater()
        legend_visible = len(self._curves) > 1

        self._plot.set_legend(legend_visible)

        return True

    def value_update(self, proxy):
        raise NotImplementedError

    def destroy_widget(self):
        """Remove all curves and their trait handlers"""
        self._curve_points = {}
        self._curves.clear()

    # ----------------------------------------------------------------
    # Plot Options

    def show_options_dialog(self):
        config = build_graph_config(self.model)
        options = config.get(KARABO_CURVE_OPTIONS, {})

        # Use model information to create a new dict[curve, option]
        curve_options = create_curve_options(
            curves=self._curves, options=options,
            curve_type=CurveType.Trend)

        dialog = CurveOptionsDialog(curve_options, parent=self.widget)
        if dialog.exec() != QDialog.Accepted:
            return

        curve_options = dialog.get_curve_options()
        if curve_options is None:
            self.reset_curve_options()
            return

        if curve_options:
            self.set_curve_options(curve_options)

    def set_curve_options(self, curve_options: dict):
        """Method to set the curve options on the model and plot"""
        self._change_model(
            {KARABO_CURVE_OPTIONS: list(curve_options.values())})
        self._plot.apply_curve_options(curve_options)

    def reset_curve_options(self):
        """Restore the default curvce options."""
        self.reset_traits(["_pens"])
        options = {}
        for proxy, curve in self._curves.items():
            pen = next(self._pens)
            options[curve] = {
                "name": proxy.key,
                "curve_type": CurveType.Trend,
                "pen_color": pen.color().name()}

        self.model.curve_options = []
        self._plot.apply_curve_options(options)

    # ----------------------------------------------------------------
    # PyQt Slots

    def _request_dialog(self):
        """Set a new time interval and request historic data"""
        x_axis = self._plot.plotItem.getAxis("bottom")
        x_min, x_max = x_axis.range
        start = QDateTime.fromSecsSinceEpoch(int(x_min))
        end = QDateTime.fromSecsSinceEpoch(int(x_max))

        dialog = RequestTimeDialog(start=start, end=end, parent=self.widget)
        if dialog.exec() == QDialog.Accepted:
            # Convert again to set the interval
            self._button_scale = False
            self._uncheck_button()
            start, end = dialog.get_start_and_end_time()
            self.set_time_interval(start, end, force=True)
            # Remove padding to avoid showing many values outside the
            # selected time-range on X-axis
            self._plot.plotItem.setRange(xRange=(start, end), padding=0)

    def _purge_curves(self):
        """Purge all curves and request historic again"""
        for v in self._curve_points.values():
            v.purge()
        x_axis = self._plot.plotItem.getAxis("bottom")
        self.set_time_interval(*x_axis.range, force=True)

    def _range_change_manually(self):
        """When the range changes manually, we stop automatically
        updating the ranges and start the timer"""
        self._timer.start()
        self._button_scale = False
        self._uncheck_button()

    def _autorange_requested(self):
        self._button_scale = False
        self._uncheck_button()
        self._plot.reset_range()

    def _change_model(self, content):
        self.model.trait_set(**restore_graph_config(content))

    def _update_intervals(self):
        """This slot is called whenever the timer timed out and previously the
        x axis scale was changed"""
        x_axis = self._plot.plotItem.getAxis("bottom")
        self.set_time_interval(*x_axis.range)

    def _update_date_widgets(self, vb, x_range):
        """This slot is called whenever the xRange changes"""
        # Note that the range comes in seconds (float)
        x_min, x_max = x_range
        start = QDateTime.fromSecsSinceEpoch(int(x_min))
        end = QDateTime.fromSecsSinceEpoch(int(x_max))

        # We change our QDateTime widgets
        self.widget.dt_start.setDateTime(start)
        self.widget.dt_end.setDateTime(end)

    def _x_axis_btns_toggled(self, button):
        """Update the x axis scale when a time button is clicked"""
        self._x_detail = self._x_axis_buttons[button]
        window_mode = self._x_detail == HIDDEN
        if window_mode:
            x_axis = self._plot.plotItem.getAxis("bottom")
            self._window = round(max(1, (x_axis.range[1] - x_axis.range[0])))
        self._window_action.setChecked(window_mode)
        # We're updating the ranges via the buttons now
        self._button_scale = True
        if self._update_axis_scale():
            self.update_later()

    # ----------------------------------------------------------------

    def deferred_update(self):
        self._update_x_range()

    def _uncheck_button(self):
        """Uncheck any checked button and the window mode"""
        self._window_action.setChecked(False)
        button_group = self.widget.bg_x_axis
        checked_button = button_group.checkedButton()
        if checked_button is not None:
            # Little workaround as we can't uncheck exclusive button group
            button_group.setExclusive(False)
            checked_button.setChecked(False)
            button_group.setExclusive(True)

    def _update_axis_scale(self):
        """The start and end time date for the x axis is updated here
        depending on the selected time button"""
        if not self._x_detail:
            return False

        if self._x_detail not in (UPTIME, HIDDEN):
            # Schedule an update for the intervals
            start_secs, end_secs = self._get_start_end_date_secs()
            # On button update we always trigger!
            self.set_time_interval(start_secs, end_secs, force=True)

        return True

    def _update_x_range(self):
        """Update x-range only in case the ranges are not regulated via mouse
        """
        if not self._button_scale:
            return

        x_range = self._get_start_end_date_secs()

        # From a deferred call the plot might have been deleted!
        if self._plot is not None:
            self._plot.plotItem.setRange(xRange=x_range)

    def _get_start_end_date_secs(self):
        """Returns the start and end date based on the selected button"""
        if self._x_detail == UPTIME:
            start = self._start_time
            end = self._get_last_timestamp()
        elif self._x_detail == HIDDEN:
            end = self._get_last_timestamp()
            start = end.addSecs(-self._window)
        else:
            start, end = get_start_end_date_time(self._x_detail)

        return start.toMSecsSinceEpoch() / 1000, end.toMSecsSinceEpoch() / 1000

    def _get_last_timestamp(self):
        """Returns the last timestamp set on the curves"""
        timestamps = []
        for curve in self._curve_points.values():
            if not curve.fill:
                continue
            timestamps.append(curve.get_max_timepoint())

        if not timestamps:
            return QDateTime.currentDateTime()
        else:
            # Preserve resolution with `fromMSecsSinceEpoch`
            # Qt has deprecated `QDateTime.fromTime_t`
            return QDateTime.fromMSecsSinceEpoch(int(max(timestamps) * 1000))

    def _draw_start_time(self, proxy, value, timestamp):
        if proxy in self._curves_start:
            self._curves_start.remove(proxy)
            start = self._start_time.toTime_t()
            if start > timestamp:
                self._curve_points[proxy].add_point(value, start)

    def set_time_interval(self, t0, t1, force=False):
        """Update the x axis scale interval of the curves"""
        for v in self._curve_points.values():
            v.changeInterval(t0, t1, force=force)


@register_binding_controller(
    ui_name="Trend Graph", klassname="DisplayTrendGraph",
    binding_type=(BoolBinding, FloatBinding, IntBinding),
    can_show_nothing=False)
class DisplayTrendGraph(BaseSeriesGraph):
    model = Instance(TrendGraphModel, args=())
    curve_type = Int(0)

    def value_update(self, proxy):
        if self.widget is None:
            return

        timestamp = proxy.binding.timestamp
        ts = timestamp.toTimestamp()
        value = proxy.value
        self._curve_points[proxy].add_point(proxy.value, ts)
        self._draw_start_time(proxy, value, timestamp=ts)


@register_binding_controller(
    ui_name="State Graph", klassname="DisplayStateGraph",
    binding_type=StringBinding,
    is_compatible=with_display_type("State"),
    can_show_nothing=False)
class DisplayStateGraph(BaseSeriesGraph):
    model = Instance(StateGraphModel, args=())
    curve_type = Int(1)

    def value_update(self, proxy):
        if self.widget is None:
            return

        timestamp = proxy.binding.timestamp
        ts = timestamp.toTimestamp()

        value = STATE_INTEGER_MAP[proxy.value]
        self._curve_points[proxy].add_point(value, ts)
        self._draw_start_time(proxy, value, timestamp=ts)


@register_binding_controller(
    ui_name="Alarm Graph", klassname="DisplayAlarmGraph",
    binding_type=StringBinding,
    is_compatible=with_display_type("AlarmCondition"),
    can_show_nothing=False)
class DisplayAlarmGraph(BaseSeriesGraph):
    model = Instance(AlarmGraphModel, args=())
    curve_type = Int(2)

    def value_update(self, proxy):
        if self.widget is None:
            return

        timestamp = proxy.binding.timestamp
        ts = timestamp.toTimestamp()

        value = ALARM_INTEGER_MAP[proxy.value]
        self._curve_points[proxy].add_point(value, ts)
        self._draw_start_time(proxy, value, timestamp=ts)



from qtpy import uic
from qtpy.QtCore import QObject, Qt, Slot
from qtpy.QtWidgets import QCompleter, QDialog, QLineEdit
from traits.api import Undefined

from karabo.native import AccessMode, Hash, create_html_hash
from karabogui import icons
from karabogui.binding.api import (
    BindingRoot, DeviceProxy, ProjectDeviceProxy, get_config_changes,
    iterate_binding)
from karabogui.singletons.api import get_topology
from karabogui.topology.api import getTopology
from karabogui.util import SignalBlocker

from .utils import get_dialog_ui


def extract_configuration(binding):
    """Extract all the values set on a binding into a Hash object. Also
    copies the reconfigurable values into another Hash."""
    def _get_binding_value(binding):
        return binding.value

    all_config = Hash()
    reconfigurable = Hash()
    for key, node in iterate_binding(binding):
        value = _get_binding_value(node)
        if value is Undefined or isinstance(value, (bytes, bytearray)):
            # Skip pipeline data
            continue
        all_config[key] = value
        if node.accessMode == AccessMode.RECONFIGURABLE:
            reconfigurable[key] = value

    return all_config, reconfigurable


class CompareDeviceConfigurationsDialog(QDialog):
    """ Dialog to show the configurations of two devices, side by side."""
    def __init__(self, reference_proxy: ProjectDeviceProxy | DeviceProxy,
                 target_proxy: ProjectDeviceProxy | DeviceProxy,
                 parent: QObject = None):
        super().__init__(parent=parent)
        self.setAttribute(Qt.WA_DeleteOnClose)
        self.setModal(False)
        ui_file = get_dialog_ui("compare_configurations.ui")
        uic.loadUi(ui_file, self)
        self.reference_device.setText(reference_proxy.device_id)
        self.target_device.setText(target_proxy.device_id)
        self._show_reference_config(reference_proxy.binding)
        self._show_target_config(target_proxy.binding)
        self._load_differences()

        self.ui_synchronize_bars.toggled.connect(
            self._synchronize_toggled)
        self._synchronize_toggled(True)
        self.ui_show_only_reconfigurable.toggled.connect(
            self._show_only_reconfigurable)

        self._show_changes = False
        self.ui_swap_button.setIcon(icons.change)
        self.ui_swap_button.clicked.connect(self._show_only_changes)
        self._show_only_changes()

    def _show_reference_config(self, binding: BindingRoot) -> None:
        all_config, reconfigurable_config = extract_configuration(binding)
        self.reference_all_config = create_html_hash(all_config)
        self.reference_configurable_config = create_html_hash(
            reconfigurable_config)
        self.reference_all = all_config
        self.reference_reconfigurable = reconfigurable_config

    def _show_target_config(self, binding: BindingRoot) -> None:
        all_config, reconfigurable_config = extract_configuration(binding)
        self.target_all_config = create_html_hash(all_config)
        self.target_reconfigurable_config = create_html_hash(
            reconfigurable_config)
        self.target_all = all_config
        self.target_reconfigurable = reconfigurable_config

    def _load_differences(self) -> None:
        """Store the difference between the configuration - full
        configurations and reconfigurable configurations separately."""
        reference_all_diff, target_all_diff = get_config_changes(
            self.reference_all, self.target_all, False)
        self.reference_all_diff = create_html_hash(reference_all_diff)
        self.target_all_diff = create_html_hash(target_all_diff)

        ref_reconf_diff, target_reconf_diff = get_config_changes(
            self.reference_reconfigurable, self.target_reconfigurable, False)
        self.reference_reconfigurable_diff = create_html_hash(ref_reconf_diff)
        self.target_reconfigurable_diff = create_html_hash(target_reconf_diff)

    def _update_text(self) -> None:
        """Update the text edit with appropriate configuration depending on
        the state of options in the gui."""
        reference = self.reference_all_config
        target = self.target_all_config
        if self._show_changes:
            if self.ui_show_only_reconfigurable.isChecked():
                reference = self.reference_reconfigurable_diff
                target = self.target_reconfigurable_diff
            else:
                reference = self.reference_all_diff
                target = self.target_all_diff
        elif self.ui_show_only_reconfigurable.isChecked():
            reference = self.reference_configurable_config
            target = self.target_reconfigurable_config

        self.reference_device_config.setHtml(reference)
        self.target_device_config.setHtml(target)

    @Slot()
    def _show_only_changes(self) -> None:
        self._show_changes = not self._show_changes
        text = "Show Configuration" if self._show_changes else "Show Changes"
        self.ui_swap_button.setText(text)
        self._update_text()

    @Slot(bool)
    def _synchronize_toggled(self, toggled: bool) -> None:
        left_scrollbar = self.reference_device_config.verticalScrollBar()
        right_scrollbar = self.target_device_config.verticalScrollBar()
        if toggled:
            left_scrollbar.valueChanged.connect(self.on_left_scrollbar_moved)
            right_scrollbar.valueChanged.connect(
                self.on_right_scrollbar_moved)
        else:
            left_scrollbar.valueChanged.disconnect(
                self.on_left_scrollbar_moved)
            right_scrollbar.valueChanged.disconnect(
                self.on_right_scrollbar_moved)

    @Slot(int)
    def on_left_scrollbar_moved(self, value: int) -> None:
        with SignalBlocker(self.target_device_config):
            self.target_device_config.verticalScrollBar().setValue(value)

    @Slot(int)
    def on_right_scrollbar_moved(self, value: int) -> None:
        with SignalBlocker(self.reference_device_config):
            self.reference_device_config.verticalScrollBar().setValue(value)

    @Slot(bool)
    def _show_only_reconfigurable(self, toggled: bool) -> None:
        self._update_text()


class DeviceSelectorDialog(QDialog):
    """Dialog to choose a device from the list of matching devices with same
    classId."""
    def __init__(self, device_id: str, parent: QObject = None):
        super().__init__(parent=parent)
        uic.loadUi(get_dialog_ui("compare_configuration_launcher.ui"), self)
        self.device_id = device_id
        text = (
            f"Select a device to compare the configuration with <span style=' "
            f"font-style:italic;'>{device_id}</span>")
        self.info_label.setText(text)
        self._load_matching_devices(device_id)

    def _load_matching_devices(self, device_id: str) -> None:
        """Populate the matching devices with same class id. """
        topology = getTopology()
        attrs = topology[f"device.{device_id}", ...]
        class_id = attrs["classId"]
        matching_devices = []

        for deviceId, _, attrs in topology["device"].iterall():
            if deviceId == self.device_id:
                # Do not show the initial device.
                continue
            classId = attrs.get("classId")
            if classId in (class_id, f"{class_id}Copy"):
                matching_devices.append(deviceId)

        if not matching_devices:
            self.info_label.setText("No devices with the same classId are "
                                    "online")
            return

        self.devices_combobox.addItems(matching_devices)
        line_edit = QLineEdit()
        completer = QCompleter(matching_devices, parent=line_edit)
        completer.setCaseSensitivity(False)
        completer.setCompletionMode(QCompleter.PopupCompletion)
        line_edit.setCompleter(completer)
        self.devices_combobox.setLineEdit(line_edit)

    def accept(self) -> None:
        topology = get_topology()
        reference = topology.get_device(self.device_id)
        target = topology.get_device(self.devices_combobox.currentText())

        def config_update():
            target.on_trait_change(config_update, "config_update", remove=True)
            target.remove_monitor()
            dialog = CompareDeviceConfigurationsDialog(
                reference_proxy=reference, target_proxy=target,
                parent=self.parent())
            dialog.exec()

        target.on_trait_change(config_update, "config_update")
        target.add_monitor()
        super().accept()
